﻿Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=2.71
@EndOfDesignText@
'From: https://www.b4x.com/android/forum/threads/class-mutablemenu-alter-menu-at-runtime.30325/
'Extended by walt61
'Class module
Sub Class_Globals

	Private MenuItems As Map
	Private mActivity As Activity
	Private mDefaultColour As Int = Colors.White
	Private mDividerColour As Int = Colors.LightGray

	Type menuItem(title As CSBuilder, eventName As String)

End Sub

' Initializes the object. You can add parameters to this method if needed.
' If defaultColour is negative, white is assumed.
' If dividerColour is negative, light grey is assumed.
Public Sub Initialize(activity As Activity, defaultColour As Int, dividerColour As Int)

	mActivity = activity
	If defaultColour >= 0 Then mDefaultColour = defaultColour
	If dividerColour >= 0 Then mDividerColour = dividerColour
	MenuItems.Initialize

End Sub

' Adds a divider with width = 'numSpaces' that get strikethrough
Public Sub AddDivider(uniqueId As String, numSpaces As Int)

	AddDivider2(uniqueId, "", numSpaces / 2)

End Sub

' Adds a divider with a String, and leading/trailing 'numSpacesBeforeAndAfter' that get strikethrough
Public Sub AddDivider2(uniqueId As String, title As String, numSpacesBeforeAndAfter As Int)

	Dim cs As CSBuilder
	AddDivider3(uniqueId, cs.Initialize.Append(title).PopAll, numSpacesBeforeAndAfter)

End Sub

' Adds a divider with a CSBuilder, and leading/trailing 'numSpacesBeforeAndAfter' that get strikethrough
Public Sub AddDivider3(uniqueId As String, title As CSBuilder, numSpacesBeforeAndAfter As Int)

	Dim sb As StringBuilder
	Dim i As Int
	Dim cs As CSBuilder

	sb.Initialize
	For i = 1 To numSpacesBeforeAndAfter
		sb.Append(" ")
	Next

	cs.Initialize.Color(mDividerColour).Strikethrough.Append(sb.ToString).Pop.Append(title).Strikethrough.Append(sb.ToString).PopAll
	AddMenuItem3(uniqueId, cs, "")

End Sub

' Adds a menu item with the given text and event name
Public Sub AddMenuItem(uniqueId As String, title As String, EventName As String)

	Dim cs As CSBuilder
	AddMenuItem3(uniqueId, cs.Initialize.Color(mDefaultColour).Append(title).PopAll, EventName)

End Sub

' Adds a menu item with the given text, text colour, and event name
Public Sub AddMenuItem2(uniqueId As String, title As String, colour As Int, EventName As String)

	Dim cs As CSBuilder
	AddMenuItem3(uniqueId, cs.Initialize.Color(colour).Append(title).PopAll, EventName)

End Sub

' Adds a menu item with the given CSBuilder and event name
Public Sub AddMenuItem3(uniqueId As String, title As CSBuilder, eventName As String)

	MenuItems.Put(uniqueId, MakeMenuItem(title, eventName))

End Sub

' Adds a menu item at a specified position (0-based) with the given text and event name
Public Sub AddMenuItemAt(uniqueId As String, title As String, eventName As String, index As Int)

	Dim cs As CSBuilder
	AddMenuItemAt3(uniqueId, cs.Initialize.Color(mDefaultColour).Append(title).PopAll, eventName, index)

End Sub

' Adds a menu item at a specified position (0-based) with the given text, text colour, and event name
Public Sub AddMenuItemAt2(uniqueId As String, title As String, colour As Int, eventName As String, index As Int)

	Dim cs As CSBuilder
	AddMenuItemAt3(uniqueId, cs.Initialize.Color(colour).Append(title).PopAll, eventName, index)

End Sub

' Adds a menu item at a specified position (0-based) with the given CSBuilder and event name
Public Sub AddMenuItemAt3(uniqueId As String, title As CSBuilder, eventName As String, index As Int)

	If index > (MenuItems.Size - 1) Then Return

	Dim n As Int
	Dim newMap As Map
	newMap.Initialize

	For n = 0 To MenuItems.Size-1
		If n = index Then newMap.Put(uniqueId, MakeMenuItem(title, eventName))
		newMap.Put(MenuItems.GetKeyAt(n), MenuItems.GetValueAt(n))		
	Next		

	MenuItems.Clear
	MenuItems = newMap

End Sub

' Updates a menu item's title with the given text
Public Sub UpdateMenuItem(uniqueId As String, title As String)

	Dim cs As CSBuilder
	UpdateMenuItem3(uniqueId, cs.Initialize.Color(mDefaultColour).Append(title).PopAll)

End Sub

' Updates a menu item's title with the given text and text colour
Public Sub UpdateMenuItem2(uniqueId As String, title As String, colour As Int)

	Dim cs As CSBuilder
	UpdateMenuItem3(uniqueId, cs.Initialize.Color(colour).Append(title).PopAll)

End Sub

' Updates a menu item's title with the given CSBuilder
Public Sub UpdateMenuItem3(uniqueId As String, title As CSBuilder)

	Dim n As Int
	Dim newMap As Map

	newMap.Initialize

	For n = 0 To MenuItems.Size-1
		If MenuItems.GetKeyAt(n) = uniqueId Then
			Dim thisMenuItem As menuItem = MenuItems.GetValueAt(n)
			newMap.Put(uniqueId, MakeMenuItem(title, thisMenuItem.eventName))
		Else
			newMap.Put(MenuItems.GetKeyAt(n), MenuItems.GetValueAt(n))
		End If
	Next

	MenuItems.Clear
	MenuItems = newMap

End Sub

' Deletes a menu item
Public Sub DeleteMenuItem(uniqueId As String)

	MenuItems.Remove(uniqueId)

End Sub

' Refreshes the menu
Public Sub RefreshMenu

	Dim n As Int
	Dim Obj1 As Reflector
	Obj1.Target = Obj1.GetActivity
	Obj1.SetField2("menuItems", Null)	

	For n = 0 To MenuItems.Size-1
		Dim thisMenuItem As menuItem = MenuItems.GetValueAt(n)
		mActivity.AddMenuItem(thisMenuItem.title, thisMenuItem.eventName)
	Next

	Obj1.RunMethod("invalidateOptionsMenu")

End Sub

' Creates a 'menuItem' type
Private Sub MakeMenuItem(title As CSBuilder, eventName As String) As menuItem

	Dim thisMenuItem As menuItem
	thisMenuItem.title = title
	thisMenuItem.eventName = eventName
	Return thisMenuItem

End Sub