﻿B4A=true
Group=Default Group
ModulesStructureVersion=1
Type=Service
Version=7.8
@EndOfDesignText@
#Region  Service Attributes 
	#StartAtBoot: False
	#ExcludeFromLibrary: True
#End Region

Sub Process_Globals
	'These global variables will be declared once when the application starts.
	'These variables can be accessed from all modules.

	' Allowed screen orientations, used in Activity_Create, as during load, we didn't want an orientation change as it could throw an exception
	Dim ORIENTATION_ANY As Int = -1
	Dim ORIENTATION_LANDSCAPE As Int = 0
	Dim ORIENTATION_PORTRAIT As Int = 1
	Dim allowedOrientation As Int = ORIENTATION_ANY ' If necessary, change this in Service_Create/CUSTOM START - customisation of skeleton variables

	Public connectedToGSMnetwork As Boolean = False

	Public clvDataList As List ' The clv data are kept in this list so that they can survive Activity_Pause

	Public handleReceivedSMSs As Boolean = False

	' Bluetooth
	Public BTadmin As BluetoothAdmin
	Private bc As ByteConverter
	Public BLEmgr As BleManager2
	Public BLEshowAll As Boolean = False
	Public BLEshowAllRaw As Boolean = False
	Public currentBLEstateText As String = "NOT INITIALISED"
	Public currentBLEstateBad As Boolean = True
	Public batteryReported As Boolean = False
	Public smsEnabled As Boolean = True
	Public BLEscanning As Boolean = False
	Public firstBLEstate As Boolean = True

	' Preferences
	Dim wmPrefsEncryptFile As Boolean = False
	Dim wmPrefsFileNameWithoutExtension As String = Application.LabelName
	Dim wmPrefsVariousStrings As Map
	Public batteryLevelAlert As Int = 20
	Public batteryLevelAlerted As Boolean = False
	Public panicAlerted As Boolean = False
	Public heartBeatMinutes As Int = 15
	Public lastHeartBeatData As String = ""
	Public lastHeartBeatDataChangedTime As Long = 0
	Public lastHeartBeatDataAlerted As Boolean = False
	Public receiveSendNotification As Boolean = False
	Public smsToNumber As String = ""
	Public smsText As String = ""

	Type puckInfo(name As String, id As String)

	Public notifIcon As Bitmap
	Public notifIconPanic As Bitmap
	Public notifID As Int = 51042 ' https://www.b4x.com/android/forum/threads/changing-notification-message.97514/#content
	Dim nb6notif As NB6

End Sub

Sub Service_Create
	'This is the program entry point.
	'This is a good place to load resources that are not specific to a single activity.


	wmUtils.wmDebugLogWrite("Starter service created")

	' Notification icons created with https://www.b4x.com/android/forum/threads/tool-make-your-own-notification-icon-with-an-ascii-and-or-a-fontawesome-material-icons-character.103754/#content
	notifIcon = LoadBitmapResize(File.DirAssets, "panicpuck.png", 24dip, 24dip, False)
	notifIconPanic = LoadBitmapResize(File.DirAssets, "panicpuck panic.png", 24dip, 24dip, False)

	wmPrefsVariousStrings.Initialize

	clvDataList.Initialize

	CallSubDelayed2(oneSms, "Initialize", Me)

	BLEmgr.Initialize("BLEmgr")
	StartService(Listener)

	StartActivity(Main)

End Sub

Sub Service_Start(StartingIntent As Intent)

End Sub

Sub Service_TaskRemoved
	'This event will be raised when the user removes the app from the recent apps list.
End Sub

'Return true to allow the OS default exceptions handler to handle the uncaught exception.
Sub Application_Error (Error As Exception, StackTrace As String) As Boolean

	Return True

End Sub

Sub Service_Destroy

	Log("Starter: Service_Destroy")
	CallSub(Listener, "StopListener")
	BLEmgr.StopScan

End Sub

Sub BLE_Scan

	Log("Starting BLE scan")
	Try
		BLEmgr.Scan2(Null, True)
		BLEscanning = True
		CallSubDelayed(Main, "SetScanMenu")
	Catch
		CallSubDelayed3(Main, "AddToClvRed", "Can't start scanning - has Bluetooth been enabled and does the app have the required permissions?", True)
	End Try

End Sub

Sub BLE_StopScan

	Log("Stopping BLE scan")
	BLEscanning = False
	Try
		BLEmgr.StopScan
		CallSubDelayed(Main, "SetScanMenu")
	Catch
		Return
	End Try

End Sub

Sub BLEmgr_StateChanged(State As Int)

	Dim subToCall As String = "AddToClvRed"
	Dim s As String

	Select State
		Case BLEmgr.STATE_POWERED_OFF
			currentBLEstateText = "POWERED OFF"
		Case BLEmgr.STATE_POWERED_ON
			currentBLEstateText = "POWERED ON"
			subToCall = "AddToClvGreen" ' We're good in this case
			currentBLEstateBad = False
		Case BLEmgr.STATE_UNSUPPORTED
			currentBLEstateText = "UNSUPPORTED"
		Case Else
			currentBLEstateText = "UNKNOWN"
	End Select

	If firstBLEstate Then
		firstBLEstate = False
		s = "BLE state is: " & currentBLEstateText
	Else
		s = "BLE state changed to: " & currentBLEstateText
	End If

	If subToCall <> "AddToClvRed" Then ' Main.AddToClvRed shows a notification, no need to do that here then
		nb6notif.Initialize("default", Application.LabelName, "LOW", True).AutoCancel(False).SmallIcon(notifIconPanic).SetDefaults(False, False, False)
		Dim notif As Notification = nb6notif.Build(Application.LabelName, s, "", Me)
		notif.Notify(notifID)
	End If

	CallSubDelayed3(Main, subToCall, s, True)

End Sub

Sub BLEmgr_DeviceFound(Name As String, Id As String, AdvertisingData As Map, RSSI As Double)

	If BLEshowAll Then
		ShowDevice_All(Name, Id, AdvertisingData, RSSI)
	Else
		ShowDevice_One(Name, Id, AdvertisingData, RSSI)
	End If

End Sub

Sub ShowDevice_One(Name As String, Id As String, AdvertisingData As Map, RSSI As Double)

	' Only handle information for the selected Puck

	Dim hex As String
	Dim splt() As String
	Dim battery As Boolean = False
	Dim panic As Boolean = False

	If Id.ToLowerCase <> Main.puckID.ToLowerCase Then Return

	wmUtils.wmLogWithTimestamp("ShowDevice_One: " & Id)

	Try
		hex = GetHex(AdvertisingData.Get(22))
		If hex <> lastHeartBeatData Then ' The broadcasted data have changed
			lastHeartBeatData = hex
			lastHeartBeatDataChangedTime = DateTime.Now
			lastHeartBeatDataAlerted = False
		Else If heartBeatMinutes > 0 Then ' Has a heartbeat (battery level broadcast) been received recently?
			If DateTime.Now > (lastHeartBeatDataChangedTime + (heartBeatMinutes * DateTime.TicksPerMinute)) Then
				CallSubDelayed3(Main, "AddToClvRed", "The Puck data haven't changed in the last " & heartBeatMinutes & " minutes, something may be wrong with the Puck", True)
				lastHeartBeatDataAlerted = True
			End If
		End If
		' Find the type of data that were broadcast: P (panic) or B (battery level)
		splt = Regex.Split("3B", hex)
		Select Case Chr(Bit.ParseInt(splt(0).SubString2(4, 6), 16))
			Case "B"
				battery = True
			Case "P"
				panic = True
			Case Else
				Return
		End Select
		Dim timestamp As String = ""
		For i = 0 To (splt(1).Length - 1) Step 2 ' Get the date and time part as yyyymmddhhmm
			timestamp = timestamp & Chr(Bit.ParseInt(splt(1).SubString2(i, i + 2), 16))
		Next
		timestamp = FormatTimeStamp(timestamp)
		If battery Then ' Battery level was broadcasted
			If panicAlerted Then CallSubDelayed3(Main, "AddToClvBlack", "Panic condition was reset", True)
			panicAlerted = False
			Dim s As String = ""
			For i = 0 To (splt(2).Length - 1) Step 2 ' Get the Puck's battery level
				s = s & Chr(Bit.ParseInt(splt(2).SubString2(i, i + 2), 16))
			Next
			Dim lvl As Int = 200 ' 200%; just in case the Puck's battery level data were corrupted
			If IsNumber(s) Then lvl = s ' Get the expected real battery data
			If batteryReported = False Then ' Report the Puck's battery level
				batteryReported = True
				CallSubDelayed3(Main, "AddToClvBlack", timestamp & CRLF & "Battery level is " & lvl & "%", True)
			End If
			If lvl < batteryLevelAlert Then ' The Puck's battery level is below the threshold set in the preferences
				If batteryLevelAlerted = False Then ' Report the Puck's battery level
					batteryLevelAlerted = True
					Log(timestamp & " - Battery level is below " & batteryLevelAlert & "%: " & lvl & "%")
					CallSubDelayed3(Main, "AddToClvRed", timestamp & CRLF & "Battery level is below " & batteryLevelAlert & "%: " & lvl & "%", True)
				End If
			End If
		Else If panic And Not(panicAlerted) Then ' Panic stations
			panicAlerted = True
			Log(timestamp & " - PANIC")
			CallSubDelayed2(Main, "AddToClvPanic", timestamp & CRLF & "PANIC")
			If smsToNumber <> "" Then SendSMS
		End If
	Catch
		Log("ShowDevice_One: " & LastException)
	End Try

End Sub

Sub ShowDevice_All(Name As String, Id As String, AdvertisingData As Map, RSSI As Double)

	' Show information about all devices that are broadcasting BLE data

	' Based on https://www.b4x.com/android/forum/threads/discovering-ibeacon-devices-with-ble2-library.60151/

	' Note: I found the device I wanted using the 'nrf connect for mobile' app
	' and looking for "0xFFFF data". The AdvertisingData looked like:
	'1: 05
	'22: FFFF423B3139373030313135323333353B313030
	'8: 5075636B
	'7: 9ECADC240EE5A9E093F3A3B50100406E
	'0: 0201051516FFFF423B3139373030313135323333353B31303005085075636B11079ECADC240EE5A9E093F3A3B50100406E00000000000000000000000000

	Dim key1 As Int = 1 ' Must be Int !
	Dim key2 As Int = 22 ' Must be Int !
	Dim val As String = "05"
	Dim sb As StringBuilder
	Dim thisPuck As puckInfo

	wmUtils.wmLogWithTimestamp("ShowDevice_All: " & Id)

	If AdvertisingData.ContainsKey(key1) = False Then Return ' Not a Puck
	If AdvertisingData.ContainsKey(key2) = False Then Return ' Not a Puck
	If bc.HexFromBytes(AdvertisingData.Get(1)) <> val Then Return ' Not a Puck

	thisPuck.Initialize
	thisPuck.name = Name
	thisPuck.id = Id

	sb.Initialize
	sb.Append("Device found: ").Append(Name).Append(" = ").Append(Id)
	If Id.ToLowerCase = Main.puckID.ToLowerCase Then sb.Append(" = your panic Puck")
	sb.Append(CRLF)

	If BLEshowAllRaw Then
		sb.Append("-------").Append("AdvertisingData:").Append("-------").Append(CRLF)
		For Each k As Int In AdvertisingData.Keys
			sb.Append($"${k}: ${bc.HexFromBytes(AdvertisingData.Get(k))}"$).Append(CRLF)
		Next
		sb.Append("-------").Append("AdvertisingData text:").Append("-------").Append(CRLF)
		For Each k As Int In AdvertisingData.Keys
			sb.Append($"${k}: ${ParseHex(GetHex(AdvertisingData.Get(k)))}"$).Append(CRLF)
		Next
	End If

	Dim hex As String = GetHex(AdvertisingData.Get(22))
	Dim splt() As String
	Dim battery As Boolean = False
	Dim panic As Boolean = False

	If BLEshowAllRaw Then
		sb.Append("-------").Append("Hex 22:").Append("-------").Append(CRLF).Append(hex).Append(CRLF).Append("-------").Append("Data:").Append("-------").Append(CRLF)
	End If

	Try
		splt = Regex.Split("3B", hex)
		Select Case Chr(Bit.ParseInt(splt(0).SubString2(4, 6), 16))
			Case "B"
				battery = True
			Case "P"
				panic = True
		End Select
		sb.Append("Time: ")
		Dim timestamp As String = ""
		For i = 0 To (splt(1).Length - 1) Step 2 ' Get the date and time part as yyyymmddhhmm
			timestamp = timestamp & Chr(Bit.ParseInt(splt(1).SubString2(i, i + 2), 16))
		Next
		sb.Append(FormatTimeStamp(timestamp))
		If battery Then
			sb.Append(CRLF).Append("Battery: ")
			For i = 0 To (splt(2).Length - 1) Step 2 ' Get the battery level
				sb.Append(Chr(Bit.ParseInt(splt(2).SubString2(i, i + 2), 16)))
			Next
			sb.Append("%")
		Else If panic Then
			sb.Append(CRLF).Append("PANIC")
		Else
			sb.Append(CRLF).Append("Not a PanicPuck")
		End If
	Catch
		Log("ShowDevice_All: " & LastException)
	End Try

	Log("====================" & CRLF & sb.ToString)
	CallSubDelayed3(Main, "AddToClvBlackWithTag", sb.ToString, thisPuck)

End Sub

Sub GetHex(b() As Byte) As String

	' Convert Byte Array to Hex

	Dim raf As RandomAccessFile
	raf.Initialize3(b, False)
	Return bc.HexFromBytes(b)

End Sub

Sub ParseHex(s As String) As String

	Dim i As Int
	Dim sb As StringBuilder

	sb.Initialize
	For i = 0 To (s.Length - 1) Step 2 ' Get the date and time part as yyyymmddhhmm
		sb.Append(Chr(Bit.ParseInt(s.SubString2(i, i + 2), 16)))
	Next

	Return sb.ToString

End Sub

Sub FormatTimeStamp(timestamp As String) As String ' timestamp: yyyymmddhhmm

	If timestamp.Length <> 12 Then
		Return(timestamp)
	Else
		Return(timestamp.SubString2(0, 4) & "-" & timestamp.SubString2(4, 6) & "-" & timestamp.SubString2(6, 8) & ";" & timestamp.SubString2(8, 10) & ":" & timestamp.SubString(10))
	End If

End Sub

Sub SendSMS

	' Panic situation; send an SMS if a number was configured in the
	' preferences and sending an SMS was enabled.

	If smsEnabled = False Then
		CallSubDelayed3(Main, "AddToClvRed", "Not sending SMS as SMS is disabled in the menu at the moment", True)
		Return
	End If

	If receiveSendNotification Then
		CallSubDelayed3(Main, "AddToClvRed", "Sending the SMS with receiveSendNotification to " & smsToNumber, True)
	Else
		CallSubDelayed3(Main, "AddToClvRed", "Sending the SMS without receiveSendNotification to " & smsToNumber, True)
	End If

	Try
		Dim Msg As typeSmsMessage
		Msg.Number = smsToNumber
		Msg.Text = smsText
		Msg.MsgID = 0
		Msg.SentReport = receiveSendNotification
		Msg.DeliveryReport = False
		CallSubDelayed2(oneSms, "SendSms", Msg)
		CallSubDelayed3(Main, "AddToClvBlack", "The SMS seems to have been sent", True)
	Catch
		CallSubDelayed3(Main, "AddToClvRed", "The SMS could Not be sent: " & LastException, True)
	End Try

End Sub

Public Sub oneSms_SendResult(Result As typeSmsResult)

	Log("oneSms_SendResult: " & Result.Success)

	If Result.Success Then
		CallSubDelayed3(Main, "AddToClvBlack", "The SMS was sent", True)
	Else
		CallSubDelayed3(Main, "AddToClvRed", "The SMS was NOT sent !!!!!", True)
	End If

End Sub