﻿B4A=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=7.3
@EndOfDesignText@
'Class module wmPreferences

'Written by walt61 - https://www.b4x.com/android/forum/members/walt61.55689/
'Version: see 'prefClassVersion' below
'Tested on Android 7 with B4A 9.00, and on Windows 10 with B4J 7.32

'IMPORTANT NOTES:
'1. If an xclv is used (see Initialize): in B4A debug mode, building up the xclv (Sub AssembleprefXCLV) is rather slow; in release mode, this is significantly faster
'2. If an xclv is used (see Initialize): the calling module should contain a 'prefsXclv_ItemClick' Sub ('prefsXclv' is then the xclv's Event Name as per the Designer) - cfr module 'ManagePreferences' in the demo app/program
'3. The class must be instantiated in each Activity that uses it as it contains Activity Objects and therefore cannot be declared in Process_Globals
'4. The B4A code uses RuntimePermissions.GetSafeDirDefaultExternal; you may need to add code to the Manifest Editor - see https://www.b4x.com/android/forum/threads/runtime-permissions-android-6-0-permissions.67689

'Dependencies:
'- B4A:
'  - Library: Core (by Erel, B4A core library)
'  - Library: Dialogs (by agraham, see https://www.b4x.com/android/forum/threads/dialogs-library.6776)
'  - Library: IME (by Erel, B4A core library)
'  - Library: RandomAccessFile (by Erel, B4A core library)
'  - Library: RuntimePermissions (by Erel, B4A core library)
'  - Library: StringUtils (by Erel, B4A core library)
'  - Library: xCustomListView (by Erel, B4J core library)
'  - Library: XUI (by Erel, B4A core library)
'- B4J:
'  - BouncyCastle, for encryption: #AdditionalJar: bcprov-jdk15on-159 (or similar)
'  - Library: CSSUtils (by Erel, B4J core library)
'  - Library: JavaObject (by Erel, B4J core library)
'  - Library: jControlsFX (by Erel, B4J core library)
'  - Library: jCore (by Erel, B4J core library)
'  - Library: jFX (by Erel, B4J core library)
'  - Library: jRandomAccessFile (by Erel, B4J core library)
'  - Library: jRLDialogsX version 1.86 or newer (by rwblinn, https://www.b4x.com/android/forum/threads/jrldialogsx-dialogs-library.52416)
'  - Library: jXUI (by Erel, B4J core library)
'  - Library: xCustomListView (by Erel, B4J core library)

'Changes:
' 00.00.03 - 2017-04-16:
' - added FontAwesome and MaterialIcons parms to AddCategory method
' - added FontAwesome and MaterialIcons parms to AddEntry method
' - internal Type changes to use CSBuilder instead of String
' - made a bunch of properties Public
' - commented Subs SetTextProperties & SetPanelProperties
' - added some properties
' - cleaned up code
' 00.00.04 - 2018-04-21:
' - changed FileDialogs' FilePath from File.DirDefaultExternal to "/mnt/sdcard"
' 00.00.05 - 2018-08-23:
' - 'prefData' is now Public instead of Private
' - added Public method 'Save'
' 00.00.06 - 2018-08-31:
' - B4J code
' - Reworked B4A code to use an xclv like the B4J code
' - Introduced 'screens'
' 00.00.07 - 2019-03-07:
' - B4A code now hides keyboard when appropriate (though this doesn't always seem to work)
' - Bug fixes
' - Passwords: show/hide (B4A) and 'repeat' (B4J)
' 00.00.08 - 2019-07-29:
' - B4A input dialogs for regular text also had a 'Show' button (which is only intended for password fields)

#Region Core

Private Sub Class_Globals

	''''''
	' Core
	''''''

	Private prefClassID As String = "wmpreferences"
	Private prefClassVersion As String = "00.00.08"
	Private prefClassLastUpdated As String = "wmpreferenceslastupdated"

	Private IamInitialized As Boolean = False
	Private overallPanelHeight As Int = 75dip

	Private allScreens As Map
	Private allCategories As Map
	Private allEntries As Map

	Private prefXCLV As CustomListView
	Private prefXCLVWidth As Double = -1
	Private jumpToStack As List

	Private screenStack As List

#If B4J
	Private fx As JFX
	Private callingForm As Form
	Private cutils As ControlsUtils
	Private theDecimalSeparator As String
	Private dirData As String
	Private lblMaterialIcons As Label
	Private lblFontawesome As Label
#End If

	Private prefXCLVassembled As Boolean = False

	Private prefFolder As String
	Private prefFile As String
	Private prefEncryptionKey As String = ""

	Public prefData As Map

#If B4A
	Dim ime1 As IME
	Dim rp As RuntimePermissions
	Type PanelType(Pnl As Panel, Color As Int, BorderWidth As Int, BorderRadius As Int, BorderColor As Int)
	Type SubPanelType(SubPnl As PanelType, Text As CSBuilder)
	Type ButtonType(Btn As Button, Color As Int, Text As CSBuilder)
	Type LabelType(Lbl As Label, Color As Int, BorderWidth As Int, BorderRadius As Int, BorderColor As Int, Text As CSBuilder)
	Type ScreenType(ID As String)
	Type CategoryType(ID As String, Pnl As PanelType, MainText As CSBuilder, SubText As CSBuilder, LongClickHint As String, Bmp As Bitmap, DependentOn As List, screen As String)
	Type EntryType(CategoryID As String, EntryID As String, CompoundID As String, Pnl As PanelType, MainText As CSBuilder, SubText As CSBuilder, LongClickHint As String, EntryInputType As Int, EntryChoices As List, DependentOn As String, defaultValues As List, fontAweSomeChar As String, materialIconsChar As String, pnlColor As Int, gotoScreen As String, passwordVisible As Boolean)
#Else If B4J
	Type PanelType(Pnl As Pane, Color As Paint, BorderWidth As Int, BorderRadius As Int, BorderColor As Paint)
	Type SubPanelType(SubPnl As PanelType, Text As String)
	Type ButtonType(Btn As Button, Color As Paint, Text As String)
	Type LabelType(lbl As Label, Color As Paint, BorderWidth As Int, BorderRadius As Int, BorderColor As Paint, Text As String)
	Type ScreenType(ID As String)
	Type CategoryType(ID As String, Pnl As PanelType, MainText As String, SubText As String, LongClickHint As String, Bmp As Image, DependentOn As List, img As Image, screen As String)
	Type EntryType(CategoryID As String, EntryID As String, CompoundID As String, Pnl As PanelType, MainText As String, SubText As String, LongClickHint As String, EntryInputType As Int, EntryChoices As List, DependentOn As String, defaultValues As List, img As Image, pnlColor As Paint, gotoScreen As String)
#End If
	Type EventAndTagType(event As String, tag As Object, xclvItemIndex As Int)

#If B4J
	Private checkbox_off_background As Image
	Private checkbox_on_background As Image
	Private back_arrow As Image
#End If

	''''''''''''''''''''''''''''''''''''
	' Appearance, configurable by caller
	''''''''''''''''''''''''''''''''''''

#If B4A
	Public categoryPanelColor As Int = Colors.LightGray
	Public categoryPanelBorderColor As Int = Colors.Black
	Public categoryMainTextColor As Int = 0xFF008080 ' Teal
	Public categorySubTextColor As Int = 0xFF008080 ' Teal
#Else If B4J
	Public categoryPanelColor As Paint = fx.Colors.LightGray
	Public categoryPanelBorderColor As Paint = fx.Colors.Black
	Public categoryMainTextColor As Paint
	Public categorySubTextColor As Paint
#End If
	Public categoryPanelBorderWidth As Int = 0
	Public categoryPanelBorderRadius As Int = 0
	Public categoryMainTextSize As Int = 20
	Public categoryMainTextBold As Boolean = False
	Public categorySubTextSize As Int = 15

#If B4A
	Public entryPanelColor As Int = Colors.White
	Public entryPanelBorderColor As Int = Colors.Black
#Else If B4J
	Public entryPanelColor As Paint = fx.Colors.White
	Public entryPanelBorderColor As Paint = fx.Colors.Black
#End If
	Public entryPanelBorderWidth As Int = 0
	Public entryPanelBorderRadius As Int = 0

#If B4A
	Public entryMainTextColor As Int = Colors.Black
	Public entrySubTextColor As Int = Colors.LightGray
	Public entryValueTextColor As Int = Colors.DarkGray
	Public entryIconsColor As Int = Colors.Blue
#Else If B4J
	Public entryMainTextColor As Paint = fx.Colors.Black
	Public entrySubTextColor As Paint = fx.Colors.LightGray
	Public entryValueTextColor As Paint = fx.Colors.DarkGray
	Public entryIconsColor As Paint = fx.Colors.Blue
#End If
	Public entryMainTextSize As Int = 20
	Public entrySubTextSize As Int = 15
	Public entryValueTextSize As Int = 15

	Private useDefaultIcons As Boolean = False
	Private ivSize As Int = 50dip
#If B4J
	Private ivTextSize As Int = 40
#End If
	Public spacingBetweenAndAroundLabels As Int = 2dip
	Public spacingBetweenXCLVpanels As Int = 1dip
	Public spacingBetweenXCLVcategories As Int = 4dip

	Private prefTitleText As String = "Preferences"
	Private prefEncryptionKeyPrompt As String = "Enter the encryption key for the preferences file"
	Private prefWrongEncryptionKey As String = "Incorrect encryption key"
	Private prefBackupText As String = ""
	Private prefRestoreText As String = ""
	Private prefResetText As String = "Reset"
	Private	prefBackupRestoreResetText As String = "Preferences"
	Private prefOKtext As String = "OK"
	Private prefCancelText As String = "Cancel"
	Private prefCompletedText As String = "Operation completed"
#If B4A
	Private prefShowText As String = "Show"
	Private prefHideText As String = "Hide"
#Else If B4J
	Private prefRepeatPwText As String = "Repeat password"
	Private prefInvalidPwText As String = "The passwords are not identical"
#End If

	Public backupCategoryLongClickText As String = ""

	''''''''''''''''''
	' Input data types
	''''''''''''''''''

	Private Const INPUT_TYPE_BACK As Int = -4 ' Reserved, do not use
	Private Const INPUT_TYPE_BACKUP As Int = -3 ' Reserved, do not use
	Private Const INPUT_TYPE_RESTORE As Int = -2 ' Reserved, do not use
	Private Const INPUT_TYPE_RESET As Int = -1 ' Reserved, do not use

	Public Const INPUT_TYPE_BOOLEAN As Int = 0 ' Entrychoices: ignored; possible values are True and False
	Public Const INPUT_TYPE_CHECKBOX As Int = 0 ' Deprecated; use INPUT_TYPE_BOOLEAN instead
	Public Const INPUT_TYPE_EDITTEXT_ANY As Int = 1 ' Entrychoices: ignored
	Public Const INPUT_TYPE_EDITTEXT_NUMBER As Int = 2 ' Entrychoices: ignored
	Public Const INPUT_TYPE_EDITTEXT_INTINCL As Int = 3 ' Entrychoices: null, or a list containing the minimum and maximum values in entries 0 and 1; both values inclusive
	Public Const INPUT_TYPE_EDITTEXT_INTEXCL As Int = 4 ' Entrychoices: null, or a list containing the minimum and maximum values in entries 0 and 1; both values exclusive
	Public Const INPUT_TYPE_INPUTLIST As Int = 5 ' Entrychoices: a list containing the valid choices
	Public Const INPUT_TYPE_INPUTMULTILIST As Int = 6 ' Entrychoices: a list containing the valid choices
	Public Const INPUT_TYPE_FOLDER As Int = 7 ' Entrychoices: ignored
	Public Const INPUT_TYPE_FILE As Int = 8 ' Entrychoices: a list containing the initial folder- and file names; null values mean 'don't offer that'
	Public Const INPUT_TYPE_DATE As Int = 9 ' Entrychoices: ignored
	Public Const INPUT_TYPE_TIME As Int = 10 ' Entrychoices: ignored
	Public Const INPUT_TYPE_DATEANDTIME As Int = 11 ' Entrychoices: ignored
#If B4A
	Public Const INPUT_TYPE_INTENT As Int = 12 ' Entrychoices: a list with 2 elements being the parameters used in 'Intent.Initialize' (Action and Uri)
#End If
	Public Const INPUT_TYPE_EDITTEXT_PASSWORD As Int = 13 ' Entrychoices: ignored
	Public Const INPUT_TYPE_SCREEN As Int = 14 ' Entrychoices: ignored
	Private Const max_input_type As Int = INPUT_TYPE_SCREEN

	''''''''''''''''''''''
	' Return status values
	''''''''''''''''''''''

	Public Const RESULT_OK As Int = 0
	Public Const RESULT_ERROR_NO_ENCRYPTION_KEY_ENTERED As Int = 1
	Public Const RESULT_ERROR_WRONG_ENCRYPTION_KEY As Int = 2
	Public Const RESULT_ERROR_ALREADY_INITIALIZED As Int = 3

	Public Const FILE_READ_OK As Int = 0
	Public Const FILE_READ_NEED_KEY As Int = 1
	Public Const FILE_READ_WRONG_KEY As Int = 2
	Public Const FILE_READ_NOT_PREFS_FILE As Int = 3
	Public Const FILE_READ_OTHER_ERROR As Int = 4

	Public Const ADDSCREEN_OK As Int = 0

	Public Const ADDCATEGORY_OK As Int = 0
	Public Const ADDCATEGORY_RESERVED_NAME As Int = 1
	Public const ADDCATEGORY_SCREEN_DOES_NOT_EXIST As Int = 2

	Public Const ADDENTRY_OK As Int = 0
	Public Const ADDENTRY_RESERVED_NAME As Int = 1
	Public Const ADDENTRY_CATEGORY_DOES_NOT_EXIST As Int = 2
	Public Const ADDENTRY_INVALID_INPUT_TYPE As Int = 3

End Sub

#If B4A
'Initializes the class:
'- xclv: the initialized CustomListView that will contain the preferences data, or Null if no user interface is required and values will be fetched and updated in code only; xclv will be made visible and populated by this class, so basically you can place it in any layout that is loaded by the calling module
'- fileNameWithoutExtension: a filename to use to store the preferences; this name will be part of the name that is eventually used
'- encryptFile: if True, the preferences file will be encrypted; if variousStrings (see below) doesn't contain 'encryptionkey' as key, the user will be prompted for the encryption key
'- hardResetByDeletingPrefsFile: if True and a preferences file already exists, it is deleted and the class works with a fresh file
'- variousStrings: a Map that (optionally) contains a number of key/value pairs (the keys are case-sensitive !) which if not specified will get default values assigned to them:
'  - key="title": the title text used in dialogs; default: "Preferences"
'  - key="encryptionkey": the encryption key to use (see aforementioned 'encryptFile' parameter)
'  - key="encryptionprompt": the prompt to be used when asking the user for the encryption key; default: "Enter the encryption key for the preferences file"
'  - key="wrongencryptionkey": the message to show if the user enters an incorrect encryption key value when restoring from backup; default: "Incorrect encryption key"
'  - key="backuprestoreresettext": the text for the backup/restore/reset category; if neither 'backup' nor 'restore' are specified, the backup/restore/reset category won't be present in the CustomListView; default: "Preferences"
'  - key="backup": the text for the 'Backup' entry; if neither 'backup' nor 'restore' are specified, the backup/restore/reset category won't be present in the CustomListView
'  - key="restore": the text for the 'Restore' entry; if neither 'backup' nor 'restore' are specified, the backup/restore/reset category won't be present in the CustomListView
'  - key="reset": the text for the 'Reset' entry; if neither 'backup' nor 'restore' are specified, the backup/restore/reset category won't be present in the CustomListView; default: "Reset"
'  NOTE: if File.ExternalWritable returns False, the backup/restore/reset options won't be available
'  - key="completed": the text that will be shown in a ToastMessage at the end of a successful backup/restore/reset operation; default: "Operation completed"
'  - key="ok": the text for the 'OK' button used in dialogs; default: "OK"
'  - key="cancel": the text for the 'Cancel' button used in dialogs; default: "Cancel"
'  - key="show": the text for the 'Show' button used in password dialogs; default: "Show"
'  - key="hide": the text for the 'Hide' button used in password dialogs; default: "Hide"
'- showDefaultIconsForEntries: if True, default icons will be shown at the left of each entry
'- Returns: property value 'RESULT_OK' if all is well, or one of the 'RESULT_ERROR_...' property values
Public Sub Initialize(xclv As CustomListView, fileNameWithoutExtension As String, encryptFile As Boolean, hardResetByDeletingPrefsFile As Boolean, variousStrings As Map, showDefaultIconsForEntries As Boolean) As Int
#Else If B4J
'Initializes the class:
'- frm: the calling form
'- xclv: the initialized CustomListView that will contain the preferences data, or Null if no user interface is required and values will be fetched and updated in code only; xclv will be made visible and populated by this class, so basically you can place it in any layout that is loaded by the calling module
'- fileNameWithoutExtension: a filename to use to store the preferences; this name will be part of the name that is eventually used
'- encryptFile: if True, the preferences file will be encrypted; if variousStrings (see below) doesn't contain 'encryptionkey' as key, the user will be prompted for the encryption key
'- hardResetByDeletingPrefsFile: if True and a preferences file already exists, it is deleted and the class works with a fresh file
'- variousStrings: a Map that (optionally) contains a number of key/value pairs (the keys are case-sensitive !) which if not specified will get default values assigned to them:
'  - key="title": the title text used in dialogs; default: "Preferences"
'  - key="encryptionkey": the encryption key to use (see aforementioned 'encryptFile' parameter)
'  - key="encryptionprompt": the prompt to be used when asking the user for the encryption key; default: "Enter the encryption key for the preferences file"
'  - key="wrongencryptionkey": the message to show if the user enters an incorrect encryption key value when restoring from backup; default: "Incorrect encryption key"
'  - key="backuprestoreresettext": the text for the backup/restore/reset category; if neither 'backup' nor 'restore' are specified, the backup/restore/reset category won't be present in the CustomListView; default: "Preferences"
'  - key="backup": the text for the 'Backup' entry; if neither 'backup' nor 'restore' are specified, the backup/restore/reset category won't be present in the CustomListView
'  - key="restore": the text for the 'Restore' entry; if neither 'backup' nor 'restore' are specified, the backup/restore/reset category won't be present in the CustomListView
'  - key="reset": the text for the 'Reset' entry; if neither 'backup' nor 'restore' are specified, the backup/restore/reset category won't be present in the CustomListView; default: "Reset"
'  - key="completed": the text that will be shown in a ToastMessage at the end of a successful backup/restore/reset operation; default: "Operation completed"
'  - key="ok": the text for the 'OK' button used in dialogs; default: "OK"
'  - key="cancel": the text for the 'Cancel' button used in dialogs; default: "Cancel"
'  - key="repeatpw": the text for the 'Repeat password' label used in dialogs; default: "Repeat password"
'  - key="invalidpw": the text for the error message used in dialogs; default: "The passwords are not identical"
'- dataDir: the default data directory to use for Backup/Restore; an empty string is allowed
'- prefsDir: the directory where the preferences file must be stored
'- callersLblMaterialIcons: an initialized MaterialIcons label that will be used to generate images if parameter 'materialIconsChr' is used in calls to 'AddEntry' or 'AddCategory'; **both** callersLblMaterialIcons and callersLblFontAwesome must be initialized Labels or both will be ignored
'- callersLblFontAwesome: an initialized FontAwesome label that will be used to generate images if parameter 'fontAwesomeChr' is used in calls to 'AddEntry' or 'AddCategory'; **both** callersLblMaterialIcons and callersLblFontAwesome must be initialized Labels or both will be ignored
'- showDefaultIconsForEntries: if True and callersLblMaterialIcons and callersLblFontAwesome are initialized, default icons will be shown at the right of each entry
'- Returns: property value 'RESULT_OK' if all is well, or one of the 'RESULT_ERROR_...' property values
Public Sub Initialize(frm As Form, xclv As CustomListView, fileNameWithoutExtension As String, encryptFile As Boolean, hardResetByDeletingPrefsFile As Boolean, variousStrings As Map, _
					dataDir As String, prefsDir As String, _
					callersLblMaterialIcons As Label, callersLblFontAwesome As Label, _
					showDefaultIconsForEntries As Boolean) As Int
#End If

	Dim s As String
	Dim s2 As String
	Dim i As Int

	If IamInitialized Then Return(RESULT_ERROR_ALREADY_INITIALIZED)

	jumpToStack.Initialize
	screenStack.Initialize
	screenStack.Add("")
	DateTime.DateFormat = DateTime.DeviceDefaultDateFormat
	DateTime.TimeFormat = DateTime.DeviceDefaultTimeFormat

#If B4A
	ime1.Initialize("")
	prefFolder = File.DirInternal
	useDefaultIcons = showDefaultIconsForEntries
#Else If B4J
	categoryMainTextColor = fx.Colors.From32Bit(0xFF008080) ' Teal
	categorySubTextColor = fx.Colors.From32Bit(0xFF008080) ' Teal

	callingForm = frm
	theDecimalSeparator = GetDecimalSeparator
	dirData = dataDir
	prefFolder = prefsDir

	If callersLblMaterialIcons.IsInitialized And callersLblFontAwesome.IsInitialized Then
		useDefaultIcons = showDefaultIconsForEntries
		lblFontawesome = callersLblFontAwesome
		lblFontawesome.Visible = False
		lblFontawesome.SetSize(ivSize, ivSize)
		lblFontawesome.TextSize = ivTextSize
		lblMaterialIcons = callersLblMaterialIcons
		lblMaterialIcons.Visible = False
		lblMaterialIcons.SetSize(ivSize, ivSize)
		lblMaterialIcons.TextSize = ivTextSize
		checkbox_off_background = GetBmpFromFontLabel(Chr(0xE835), entryIconsColor, lblMaterialIcons, entryPanelColor)
		checkbox_on_background = GetBmpFromFontLabel(Chr(0xE834), entryIconsColor, lblMaterialIcons, entryPanelColor)
		back_arrow = GetBmpFromFontLabel(Chr(0xF060), categoryMainTextColor, lblFontawesome, categoryPanelColor)
	End If
#End If

	If variousStrings.ContainsKey("encryptionkey") Then prefEncryptionKey = variousStrings.Get("encryptionkey")
	If variousStrings.ContainsKey("encryptionprompt") Then prefEncryptionKeyPrompt = variousStrings.Get("encryptionprompt")

	s = ""
	For i = 0 To (fileNameWithoutExtension.Length - 1)
		s2 = fileNameWithoutExtension.SubString2(i, i + 1).ToLowerCase
		If IsNumber(s2) Then
			s = s & s2
		Else If (s2.CompareTo("a") >= 0) And (s2.CompareTo("z") <= 0) Then
			s = s & s2
		End If
	Next
	prefFile = s & "preferences.dat"

	If hardResetByDeletingPrefsFile Then
		If File.Exists(prefFolder, prefFile) Then File.Delete(prefFolder, prefFile)
	End If

	If encryptFile And (prefEncryptionKey = "") Then
#If B4A ' In B4J, this for some reason affects (badly...) the layout of the preferences form and makes the icons narrower
		prefEncryptionKey = GetEncryptionKey("")
#End If
		If prefEncryptionKey = "" Then Return(RESULT_ERROR_NO_ENCRYPTION_KEY_ENTERED)
	End If

	If File.Exists(prefFolder, prefFile) Then
		If ReadPrefsFile(prefEncryptionKey, prefFolder, prefFile) <> FILE_READ_OK Then Return(RESULT_ERROR_WRONG_ENCRYPTION_KEY)
	Else
		prefData = CreateMap(prefClassID:prefClassVersion)
	End If

	allScreens.Initialize
	allScreens.Put("", "")
	allCategories.Initialize
	allEntries.Initialize

	If xclv <> Null Then
		If xclv.IsInitialized Then
			prefXCLV = xclv
#If B4A
			prefXCLV.AsView.Color = Colors.LightGray
#Else If B4J
			prefXCLV.AsView.Color = fx.Colors.To32Bit(fx.Colors.LightGray)
#End If
		End If
	End If

	If variousStrings.ContainsKey("title") Then prefTitleText = variousStrings.Get("title")
	If variousStrings.ContainsKey("backup") Then prefBackupText = variousStrings.Get("backup")
	If variousStrings.ContainsKey("restore") Then prefRestoreText = variousStrings.Get("restore")
	If variousStrings.ContainsKey("reset") Then prefResetText = variousStrings.Get("reset")
	If variousStrings.ContainsKey("backuprestoreresettext") Then prefBackupRestoreResetText = variousStrings.Get("backuprestoreresettext")
	If variousStrings.ContainsKey("ok") Then prefOKtext = variousStrings.Get("ok")
	If variousStrings.ContainsKey("cancel") Then prefCancelText = variousStrings.Get("cancel")
	If variousStrings.ContainsKey("completed") Then prefCompletedText = variousStrings.Get("completed")
#If B4A
	If variousStrings.ContainsKey("show") Then prefShowText = variousStrings.Get("show")
	If variousStrings.ContainsKey("hide") Then prefHideText = variousStrings.Get("hide")
#Else If B4J
	If variousStrings.ContainsKey("repeatpw") Then prefRepeatPwText = variousStrings.Get("repeatpw")
	If variousStrings.ContainsKey("invalidpw") Then prefInvalidPwText = variousStrings.Get("invalidpw")
#End If

	If (prefBackupText = "") And (prefRestoreText <> "") Then
		prefBackupText = "Backup..."
	Else If (prefBackupText <> "") And (prefRestoreText = "") Then
		prefRestoreText = "Restore..."
	End If

	If xclv <> Null Then
		If xclv.IsInitialized Then
			If (prefBackupText <> "") Or (prefRestoreText <> "") Then
#If B4A
				If File.ExternalWritable Then
#End If
					AddCategoryPrivate("backuprestorereset", "", prefBackupRestoreResetText, "", backupCategoryLongClickText, Null, False, Null, True, Chr(0xF1CD), "")
					AddEntryPrivate("backuprestorereset", "bkp", prefBackupText, "", "", INPUT_TYPE_BACKUP, Null, Null, "", Chr(0xF0C7), "", "", True)
					AddEntryPrivate("backuprestorereset", "rstr", prefRestoreText, "", "", INPUT_TYPE_RESTORE, Null, Null, "", Chr(0xF07C), "", "", True)
					AddEntryPrivate("backuprestorereset", "reset", prefResetText, "", "", INPUT_TYPE_RESET, Null, Null, "", "", Chr(0xE8BA), "", True)
				End If
#If B4A
			End If
#End If
		End If
	End If

	IamInitialized = True
	Return(RESULT_OK)

End Sub

' Refreshes the values from the file, e.g. after the user (potentially) changed values
' via the app's preference manager activity. If the class wasn't initialized, nothing
' is done and the method just returns.
Public Sub Refresh

	If IamInitialized Then ReadPrefsFile(prefEncryptionKey, prefFolder, prefFile)

End Sub

#End Region ' Core

#Region ShowHide

'Makes the CustomListView with the preferences visible. Must be called after Initialize as Initialize won't show the CustomListView.
'Does nothing if the user interface was not enabled in the Initialize call.
Public Sub Show

	If prefXCLV <> Null Then
		If prefXCLV.IsInitialized Then
			If prefXCLVassembled = False Then AssembleprefXCLV
			jumpToStack.Initialize
			prefXCLV.AsView.Visible = True
		End If
	End If

End Sub

' Expands all categories; should only be called after all categories have been added
Public Sub ExpandAllCategories

	For Each oneCat As CategoryType In allCategories.Values
		prefData.Put(oneCat.ID & "iscollapsed", False)
	Next

	WritePrefsFile(prefEncryptionKey, False, "", "")
	AssembleprefXCLV

End Sub

' Collapses all categories; should only be called after all categories have been added
Public Sub CollapseAllCategories

	For Each oneCat As CategoryType In allCategories.Values
		prefData.Put(oneCat.ID & "iscollapsed", True)
	Next

	WritePrefsFile(prefEncryptionKey, False, "", "")
	AssembleprefXCLV

End Sub

Private Sub AssembleprefXCLV

	Dim totalHeightWithinPanel As Int
	Dim currentScreen As String = ""

	If prefXCLVWidth < 0 Then prefXCLVWidth = prefXCLV.AsView.Width ' No idea why prefXCLV.AsView.Width had changed as of the 2nd time this Sub was called, so I worked around that by using this variable

	prefXCLV.AsView.Visible = False
	prefXCLV.Clear
	prefXCLV.AsView.Width = prefXCLVWidth
	prefXCLVassembled = True

	If screenStack.Size > 1 Then
		currentScreen = screenStack.Get(0)
		AssembleprefXCLVBackEntry
	End If

	' Categories
	For Each oneCat As CategoryType In allCategories.Values

		Dim showIt As Boolean = True

		If oneCat.screen <> currentScreen Then
			showIt = False
		Else If oneCat.DependentOn <> Null Then
			If oneCat.DependentOn.Size = 2 Then
				Try
					Dim depCat As String = oneCat.DependentOn.Get(0)
					depCat = depCat.Trim.ToLowerCase
					If depCat <> oneCat.ID Then
						Dim depEnt As String = oneCat.DependentOn.Get(1)
						showIt = ConvertToBoolean(prefData.Get(depEnt.Trim.ToLowerCase))
					End If
				Catch
					showIt = False ' Dependency value not found (or entry isn't of type 'INPUT_TYPE_BOOLEAN'), so consider this 'not checked'
				End Try
			End If
		End If

		If showIt Then
			Dim ett As EventAndTagType
			ett.event = "Category"
			ett.tag = oneCat
			ett.xclvItemIndex = prefXCLV.Size
#If B4A
			Dim categoryPnl As Panel
			categoryPnl.Initialize("Category")
			categoryPnl.Tag = ett
			categoryPnl.Color = oneCat.Pnl.Color
			totalHeightWithinPanel = AddXCLVlabel(True, categoryPnl, ett, spacingBetweenAndAroundLabels, prefXCLVWidth - 2dip, oneCat.MainText, categoryMainTextColor, categoryMainTextSize, categoryMainTextBold, Null)
#Else If B4J
			Dim categoryPnl As Pane
			categoryPnl.Initialize("Category")
			categoryPnl.Tag = ett
			CSSUtils.SetBackgroundColor(categoryPnl, oneCat.Pnl.Color)
			totalHeightWithinPanel = AddXCLVlabel(True, categoryPnl, ett, spacingBetweenAndAroundLabels, prefXCLVWidth - 2dip, oneCat.MainText, categoryMainTextColor, categoryMainTextSize, categoryMainTextBold, oneCat.img)
#End If
			If oneCat.SubText.Length > 0 Then
				totalHeightWithinPanel = AddXCLVlabel(True, categoryPnl, ett, totalHeightWithinPanel, prefXCLVWidth - 2dip, oneCat.SubText, categorySubTextColor, categorySubTextSize, False, Null)
			End If
			AddXCLVpanel(categoryPnl, totalHeightWithinPanel, oneCat.Pnl.Color, ett)

			If CategoryIsCollapsed(oneCat) = False Then
				For Each oneEntry As EntryType In allEntries.Values
					AssembleprefXCLVentries(oneEntry, oneCat)
				Next
			End If
		End If ' showIt

	Next

	Try
		' Ensure all xclv entries will be completely visible
		' With less than 100dip, the last entry wasn't fully visible on a Moto-X phone
		prefXCLV.sv.ScrollViewInnerPanel.Height = prefXCLV.sv.ScrollViewInnerPanel.Height + 100dip
	Catch
		Log(LastException)
	End Try

	Try
		If jumpToStack.Size > 0 Then
			Sleep(0)
			prefXCLV.JumpToItem(jumpToStack.Get(0))
			jumpToStack.RemoveAt(0)
		End If
	Catch
		Log("AssembleprefXCLV: " & LastException)
	End Try
	prefXCLV.AsView.Visible = True

End Sub

Private Sub AssembleprefXCLVBackEntry

	Dim totalHeightWithinPanel As Int
	Dim lblWidth As Double = prefXCLVWidth - ivSize - 4dip - 12dip ' 12dip for the verttical scrollbar
	Dim oneEntry As EntryType
	Dim ett As EventAndTagType

	oneEntry.Initialize
	oneEntry.EntryInputType = INPUT_TYPE_BACK

	ett.event = "Entry"
	ett.tag = oneEntry
	ett.xclvItemIndex = prefXCLV.Size

	Dim iv As ImageView
	iv.Initialize("Entry")
	iv.Tag = ett

#If B4A
	Dim entryPnl As Panel
	entryPnl.Initialize("Entry")
	entryPnl.Tag = ett
	entryPnl.Color = categoryPanelColor
	entryPnl.AddView(iv, lblWidth + 2dip, 0, ivSize, ivSize)
	iv.Gravity = Gravity.FILL
	Dim cs As CSBuilder
	cs.Initialize.Color(categoryMainTextColor).size(entryMainTextSize).Typeface(Typeface.FONTAWESOME).Append(Chr(0xF060)).PopAll
	totalHeightWithinPanel = AddXCLVlabel(False, entryPnl, ett, spacingBetweenAndAroundLabels, lblWidth, cs, categoryMainTextColor, categoryMainTextSize, False, Null)
#Else If B4J
	Dim entryPnl As Pane
	entryPnl.Initialize("Entry")
	entryPnl.Tag = ett
	CSSUtils.SetBackgroundColor(entryPnl, categoryPanelColor)
	entryPnl.AddNode(iv, 0, 0, ivSize, ivSize)
	iv.SetImage(back_arrow)
	totalHeightWithinPanel = AddXCLVlabel(False, entryPnl, ett, spacingBetweenAndAroundLabels, lblWidth, oneEntry.MainText, categoryMainTextColor, categoryMainTextSize, False, oneEntry.img)
#End If

	totalHeightWithinPanel = Max(totalHeightWithinPanel, ivSize)
	AddXCLVpanel(entryPnl, totalHeightWithinPanel, categoryPanelColor, ett)

End Sub

#If B4A
Private Sub AssembleprefXCLVentries(oneEntry As EntryType, oneCat As CategoryType)
#Else If B4J
Private Sub AssembleprefXCLVentries(oneEntry As EntryType, oneCat As CategoryType)
#End If
	Dim totalHeightWithinPanel As Int
	Dim lblWidth As Double = prefXCLVWidth - ivSize - 4dip - 12dip ' 12dip for the verttical scrollbar
	Dim ett As EventAndTagType
	Dim curValue As String = ""

	If oneEntry.CategoryID <> oneCat.ID Then Return

	If oneEntry.DependentOn <> "" Then
		If ConvertToBoolean(prefData.Get(oneEntry.DependentOn)) = False Then Return
	End If

	ett.event = "Entry"
	ett.tag = oneEntry
	ett.xclvItemIndex = prefXCLV.Size

	Select Case oneEntry.EntryInputType
		Case INPUT_TYPE_BOOLEAN, INPUT_TYPE_BACKUP, INPUT_TYPE_RESTORE, INPUT_TYPE_RESET, INPUT_TYPE_BACK
			' Do nothing
		Case Else
			curValue = GetFormattedData(oneEntry)
	End Select

	Dim iv As ImageView
	iv.Initialize("Entry")
	iv.Tag = ett

#If B4A
	Dim entryPnl As Panel
	Dim frontPadding As Boolean = False
	entryPnl.Initialize("Entry")
	entryPnl.Tag = ett
	entryPnl.Color = oneEntry.Pnl.Color
	entryPnl.AddView(iv, lblWidth + 2dip, 0, ivSize, ivSize)
	iv.Gravity = Gravity.FILL
	Dim cs As CSBuilder
	Select Case oneEntry.EntryInputType
		Case INPUT_TYPE_BOOLEAN
			frontPadding = True
			Try
				If prefData.Get(oneEntry.EntryID) Then
					cs.Initialize.Color(entryMainTextColor).size(entryMainTextSize).Typeface(Typeface.MATERIALICONS).Append(Chr(0xE834)).Pop.Append(" ").Append(oneEntry.MainText).PopAll
				Else
					cs.Initialize.Color(entryMainTextColor).size(entryMainTextSize).Typeface(Typeface.MATERIALICONS).Append(Chr(0xE835)).Pop.Append(" ").Append(oneEntry.MainText).PopAll
				End If
			Catch
				cs.Initialize.Color(entryMainTextColor).size(entryMainTextSize).Typeface(Typeface.MATERIALICONS).Append(Chr(0xE835)).Pop.Append(" ").Append(oneEntry.MainText).PopAll
			End Try
		Case Else
			cs = oneEntry.MainText
	End Select
	totalHeightWithinPanel = AddXCLVlabel(False, entryPnl, ett, spacingBetweenAndAroundLabels, lblWidth, cs, entryMainTextColor, entryMainTextSize, False, Null)
	If oneEntry.SubText.Length > 0 Then
		If frontPadding Then ' Add a checkbox item in the same colour as the panel so that it's not visible, while the text will be aligned properly with MainText
			cs.Initialize.Color(entryPanelColor).size(entryMainTextSize).Typeface(Typeface.MATERIALICONS).Append(Chr(0xE835)).Popall.Color(entrySubTextColor).Size(entrySubTextSize).Append(" ").Append(oneEntry.SubText).PopAll
			totalHeightWithinPanel = AddXCLVlabel(False, entryPnl, ett, totalHeightWithinPanel, lblWidth, cs, entrySubTextColor, entrySubTextSize, False, Null)
		Else
			totalHeightWithinPanel = AddXCLVlabel(False, entryPnl, ett, totalHeightWithinPanel, lblWidth, oneEntry.SubText, entrySubTextColor, entrySubTextSize, False, Null)
		End If
	End If
	Select Case curValue
		Case "", Null, "null"
			' Do nothing
		Case Else
			Dim csCurValue As CSBuilder
			If oneEntry.fontAweSomeChar <> "" Then
				csCurValue.Initialize.Size(entryMainTextSize).Color(oneEntry.pnlColor).Typeface(Typeface.FONTAWESOME).Append(oneEntry.fontAweSomeChar).Pop.Append(" ").PopAll.Size(entryValueTextSize).Color(entryValueTextColor).Append(curValue).PopAll
			Else If oneEntry.materialIconsChar <> "" Then
				csCurValue.Initialize.Size(entryMainTextSize).Color(oneEntry.pnlColor).Typeface(Typeface.MATERIALICONS).Append(oneEntry.materialIconsChar).Pop.Append(" ").PopAll.Size(entryValueTextSize).Color(entryValueTextColor).Append(curValue).PopAll
			Else
				csCurValue.Initialize.Color(entryValueTextColor).size(entryValueTextSize).Append(curValue).PopAll
			End If
			totalHeightWithinPanel = AddXCLVlabel(False, entryPnl, ett, totalHeightWithinPanel, lblWidth, csCurValue, entryValueTextColor, entryMainTextSize, False, Null)
	End Select
#Else If B4J
	Dim entryPnl As Pane
	entryPnl.Initialize("Entry")
	entryPnl.Tag = ett
	CSSUtils.SetBackgroundColor(entryPnl, oneEntry.Pnl.Color)
	entryPnl.AddNode(iv, lblWidth + 2dip, 0, ivSize, ivSize)
	If oneEntry.img <> Null Then
		If oneEntry.img.IsInitialized Then iv.SetImage(oneEntry.img)
	End If
	Select Case oneEntry.EntryInputType
		Case INPUT_TYPE_BOOLEAN
			Try
				If prefData.Get(oneEntry.EntryID) Then
					iv.SetImage(checkbox_on_background)
				Else
					iv.SetImage(checkbox_off_background)
				End If
			Catch
				iv.SetImage(checkbox_off_background)
			End Try
	End Select
	totalHeightWithinPanel = AddXCLVlabel(False, entryPnl, ett, spacingBetweenAndAroundLabels, lblWidth, oneEntry.MainText, entryMainTextColor, entryMainTextSize, False, oneEntry.img)
	If oneEntry.SubText.Length > 0 Then
		totalHeightWithinPanel = AddXCLVlabel(False, entryPnl, ett, totalHeightWithinPanel, lblWidth, oneEntry.SubText, entrySubTextColor, entrySubTextSize, False, Null)
	End If
	Select Case curValue
		Case "", Null, "null"
	' Do nothing
		Case Else
			totalHeightWithinPanel = AddXCLVlabel(False, entryPnl, ett, totalHeightWithinPanel, lblWidth, curValue, entryValueTextColor, entryMainTextSize, False, Null)
	End Select
#End If

	totalHeightWithinPanel = Max(totalHeightWithinPanel, ivSize)
	AddXCLVpanel(entryPnl, totalHeightWithinPanel, oneEntry.Pnl.Color, ett)

End Sub

#If B4A
Private Sub AddXCLVlabel(isCategory As Boolean, parentPnl As Panel, ett As EventAndTagType, top As Int, width As Int, _
						text As CSBuilder, textColour As Int, textSize As Int, textBold As Boolean, img As Bitmap) As Int
#Else If B4J
Private Sub AddXCLVlabel(isCategory As Boolean, parentPnl As Pane, ett As EventAndTagType, top As Int, width As Double, _
						text As String, textColour As Paint, textSize As Int, textBold As Boolean, img As Image) As Int
#End If
	Dim lbl As Label
	lbl.Initialize(ett.event)

#If B4A
	Dim su As StringUtils
	lbl.SingleLine = False
	lbl.Tag = ett
	lbl.Text = text
	If textSize >= 0 Then lbl.TextSize = textSize
	If textColour >= 0 Then lbl.TextColor = textColour
	If textBold Then lbl.Typeface = Typeface.DEFAULT_BOLD
	parentPnl.AddView(lbl, 2dip, top, width, 20dip)
	lbl.Height = su.MeasureMultilineTextHeight(lbl, lbl.Text) + spacingBetweenAndAroundLabels ' Adding 'spacingBetweenAndAroundLabels' as otherwise some labels' height was insufficient and the text was truncated at the bottom
	lbl.Gravity = Gravity.LEFT + Gravity.CENTER_VERTICAL
	Return(top + lbl.Height + spacingBetweenAndAroundLabels)
#Else If B4J
	Dim iv As ImageView
	Dim lblLeft As Int = 2dip
	Dim margin As Int = lblLeft
	Dim lblHeight As Double
	Dim hasIv As Boolean = False

	lbl.WrapText = True ' Notice: text still gets ellipsized if it doesn't contain spaces!
	If textSize >= 0 Then lbl.TextSize = textSize
	lbl.TextColor = textColour

	If isCategory Then
		If img <> Null Then
			If img.IsInitialized Then
				lblLeft = (2 * margin) + ivSize
				hasIv = True
				width = width - lblLeft
				iv.Initialize(ett.event)
				parentPnl.AddNode(iv, margin, top, ivSize, ivSize)
				iv.SetImage(img)
			End If
		End If
	End If

	parentPnl.AddNode(lbl, lblLeft, top, width, 20dip)
	lbl.Tag = ett
	lbl.Text = text
	If textBold Then CSSUtils.SetStyleProperty(lbl, "-fx-font-weight", "bold")
	lblHeight = MeasureMultilineTextHeight(fx.DefaultFont(lbl.TextSize), width, lbl.Text)
	lbl.SetSize(width, lblHeight)
	CSSUtils.SetStyleProperty(lbl, "-fx-alignment", "CENTER-LEFT")

	If hasIv And (lblHeight < ivSize) Then
		Return(top + ivSize + spacingBetweenAndAroundLabels)
	Else
		Return(top + lblHeight + spacingBetweenAndAroundLabels)
	End If
#End If

End Sub

#If B4A
Private Sub AddXCLVpanel(pnlIn As Panel, pnlHeight As Int, pnlColor As Int, tag As EventAndTagType)
#Else If B4J
Private Sub AddXCLVpanel(pnlIn As Pane, pnlHeight As Int, pnlColor As Paint, tag As EventAndTagType)
#End If

	Dim pnlOutHeight As Int

#If B4A
	Dim pnlOut As Panel
	pnlOut.Initialize("")
	pnlOut.Color = pnlColor
	If overallPanelHeight >= pnlHeight Then
		pnlOut.AddView(pnlIn, 0, (overallPanelHeight - pnlHeight) / 2, prefXCLVWidth, pnlHeight)
		pnlOutHeight = overallPanelHeight
	Else
		pnlOut.AddView(pnlIn, 0, 0, prefXCLVWidth, pnlHeight)
		pnlOutHeight = pnlHeight
	End If
	pnlOut.Width = prefXCLVWidth
	pnlOut.Height = pnlOutHeight
#Else If B4J
	Dim pnlOut As Pane
	pnlOut.Initialize("")
	CSSUtils.SetBackgroundColor(pnlOut, pnlColor)
	If overallPanelHeight >= pnlHeight Then
		pnlOut.AddNode(pnlIn, 0, (overallPanelHeight - pnlHeight) / 2, prefXCLVWidth, pnlHeight)
		pnlOutHeight = overallPanelHeight
	Else
		pnlOut.AddNode(pnlIn, 0, 0, prefXCLVWidth, pnlHeight)
		pnlOutHeight = pnlHeight
	End If
	pnlOut.SetSize(prefXCLVWidth, pnlOutHeight)
#End If

	prefXCLV.Add(pnlOut, tag)

End Sub

Private Sub GetFormattedData(oneEntry As EntryType) As String

	Try
		If prefData.Get(oneEntry.EntryID) = Null Then Return("")
	Catch
		Log("") ' No action needed
	End Try

	Try
		Select Case oneEntry.EntryInputType
			Case INPUT_TYPE_DATE
				Return(DateTime.Date(prefData.Get(oneEntry.EntryID)))
			Case INPUT_TYPE_TIME
				Return(DateTime.Time(prefData.Get(oneEntry.EntryID)))
			Case INPUT_TYPE_DATEANDTIME
				Dim currentSelections As List
				currentSelections.Initialize
				currentSelections.AddAll(prefData.get(oneEntry.EntryID))
				Return(DateTime.Date(currentSelections.Get(0)) & " - " & DateTime.Time(currentSelections.Get(1)))
			Case INPUT_TYPE_BOOLEAN, INPUT_TYPE_SCREEN
				Return("")
#If B4A
			Case INPUT_TYPE_INTENT
				Return("")
#End If
			Case INPUT_TYPE_INPUTMULTILIST
				Dim i As Int
				Dim s As String = ""
				Dim currentSelections As List
				currentSelections.Initialize
				currentSelections.AddAll(prefData.get(oneEntry.EntryID))
				For i = 0 To (currentSelections.Size - 1)
					If s <> "" Then s = s & " | "
					s = s & currentSelections.Get(i)
				Next
				Return(s)
			Case INPUT_TYPE_EDITTEXT_PASSWORD
				If prefData.Get(oneEntry.EntryID) = "" Then
					Return ""
				Else
					Return "********"
				End If
			Case Else
				Return(prefData.Get(oneEntry.EntryID))
		End Select
	Catch
		Return("")
	End Try

End Sub

'Hides the xCustomListView containing the preferences
Public Sub Hide

	If prefXCLV.IsInitialized Then prefXCLV.AsView.Visible = False

End Sub

#End Region ' ShowHide

#Region Setup

'Adds a preference screen; only relevant when using the user interface:
'- ID: a unique case-insensitive ID that will be used internally; the ID of the default screen that is automatically added by Initialize is an empty string
Public Sub AddScreen(ID As String) As Int

	If ID.Trim <> "" Then
		Dim newScreen As ScreenType
		newScreen.ID = ID.Trim.ToLowerCase
		allScreens.Put(newScreen.ID, newScreen)
	End If
	Return(ADDSCREEN_OK)

End Sub

#If B4A
'Adds a preference category; only required when using the user interface:
'- ID: a unique case-insensitive ID that will be used internally; do not use 'BackupRestore' (case insensitive) as category ID (the category won't be added then)
'- screen: the screen to which the category belongs; it must have been defined with AddScreen, or use an empty string for the root screen
'- mainText: the category's title
'- subText: a text to display underneath mainText in the category panel in the preferences CustomListView
'- longClickHint: the text to display when the user long-clicks the category
'- bmp: Null, or a bitmap to show in the category panel in the preferences CustomListView; presently not used
'- initiallyCollapsed: if True, the category will be initially collapsed and can be expanded by clicking it; after the first time that AddCategory is called (ever), the category's collapsed/expanded state will be stored in/retrieved from the preferences file
'- dependentOn: Null, or a List containing the category ID and entry ID in elements 0 and 1 respectively, of an entry (from another category) of type 'INPUT_TYPE_BOOLEAN'; if that checkbox is checked, the new category will be shown; when it is unchecked, the category will be hidden
'- fontAwesomeChr: "" or the Chr() value of the FontAwesome character to use in front of mainText
'- materialIconsChr: "" or the Chr() value of the MaterialIcons character to use in front of mainText
'- Returns: ADDCATEGORY_OK if all went well, or ADDCATEGORY_RESERVED_NAME if a reserved name was used
Public Sub AddCategory(ID As String, screen As String, mainText As String, subText As String, longClickHint As String, bmp As Bitmap, initiallyCollapsed As Boolean, dependentOn As List, fontAwesomeChr As String, materialIconsChr As String) As Int
#Else If B4J
'Adds a preference category; only required when using the user interface:
'- ID: a unique case-insensitive ID that will be used internally; do not use 'BackupRestore' (case insensitive) as category ID (the category won't be added then)
'- screen: the screen to which the category belongs; it must have been defined with AddScreen, or use an empty string for the root screen
'- mainText: the category's title
'- subText: a text to display underneath mainText in the category panel in the preferences CustomListView
'- longClickHint: the text to display when the user long-clicks the category
'- bmp: Null, or a bitmap to show in the category panel in the preferences CustomListView; presently not used
'- initiallyCollapsed: if True, the category will be initially collapsed and can be expanded by clicking it; after the first time that AddCategory is called (ever), the category's collapsed/expanded state will be stored in/retrieved from the preferences file
'- dependentOn: Null, or a List containing the category ID and entry ID in elements 0 and 1 respectively, of an entry (from another category) of type 'INPUT_TYPE_BOOLEAN'; if that checkbox is checked, the new category will be shown; when it is unchecked, the category will be hidden
'- fontAwesomeChr: "" or the Chr() value of the FontAwesome character to use in front of mainText
'- materialIconsChr: "" or the Chr() value of the MaterialIcons character to use in front of mainText
'- Returns: ADDCATEGORY_OK if all went well, or ADDCATEGORY_RESERVED_NAME if a reserved name was used
Public Sub AddCategory(ID As String, screen As String, mainText As String, subText As String, longClickHint As String, bmp As Image, initiallyCollapsed As Boolean, dependentOn As List, fontAwesomeChr As String, materialIconsChr As String) As Int
#End If

	Return(AddCategoryPrivate(ID, screen, mainText, subText, longClickHint, bmp, initiallyCollapsed, dependentOn, False, fontAwesomeChr, materialIconsChr))

End Sub

#If B4A
Private Sub AddCategoryPrivate(ID As String, screen As String, mainText As String, subText As String, longClickHint As String, bmp As Bitmap, initiallyCollapsed As Boolean, dependentOn As List, allowReserved As Boolean, fontAwesomeChr As String, materialIconsChr As String) As Int

	Dim pnl As Panel
	Dim mainTextVar As CSBuilder
	Dim subTextVar As CSBuilder
#Else If B4J
Private Sub AddCategoryPrivate(ID As String, screen As String, mainText As String, subText As String, longClickHint As String, bmp As Image, initiallyCollapsed As Boolean, dependentOn As List, allowReserved As Boolean, fontAwesomeChr As String, materialIconsChr As String) As Int
	Dim pnl As Pane
	Dim mainTextVar As String
	Dim subTextVar As String
#End If

	Dim newCatPnl As PanelType
	Dim newCat As CategoryType

	If allowReserved = False Then
		Select Case ID.Trim.ToLowerCase
			Case "backuprestore"
				Return(ADDCATEGORY_RESERVED_NAME) ' Reserved name
		End Select
	End If

	If screen.Trim <> "" Then
		If allScreens.ContainsKey(screen.Trim.ToLowerCase) = False Then Return(ADDCATEGORY_SCREEN_DOES_NOT_EXIST)
	End If

	pnl.Initialize("Category")

	newCatPnl.Pnl = pnl
	newCatPnl.Color = categoryPanelColor
	newCatPnl.BorderWidth = categoryPanelBorderWidth
	newCatPnl.BorderRadius = categoryPanelBorderRadius
	newCatPnl.BorderColor = categoryPanelBorderColor

#If B4A
	If fontAwesomeChr <> "" Then
		mainTextVar.Initialize.Color(categoryMainTextColor).size(categoryMainTextSize).Typeface(Typeface.FONTAWESOME).Append(fontAwesomeChr).Append(" ").Typeface(Typeface.DEFAULT).Append(mainText).PopAll
	Else If materialIconsChr <> "" Then
		mainTextVar.Initialize.Color(categoryMainTextColor).size(categoryMainTextSize).Typeface(Typeface.MATERIALICONS).Append(materialIconsChr).Append(" ").Typeface(Typeface.DEFAULT).Append(mainText).PopAll
	Else
		mainTextVar.Initialize.Color(categoryMainTextColor).size(categoryMainTextSize).Append(mainText).PopAll
	End If
	If subText = "" Then
		subTextVar.Initialize
	Else
		subTextVar.Initialize.Color(categorySubTextColor).size(categorySubTextSize).Append(subText).PopAll
	End If
#Else If B4J
	mainTextVar = mainText
	subTextVar = subText
#End If

	newCat.ID = ID.Trim.ToLowerCase
	newCat.pnl = newCatPnl
	newCat.MainText = mainTextVar
	newCat.SubText = subTextVar
	newCat.LongClickHint = longClickHint.Trim
	newCat.screen = screen.trim.ToLowerCase
	newCat.Bmp = bmp
#If B4J
	If fontAwesomeChr <> "" Then newCat.img = GetBmpFromFontLabel(fontAwesomeChr, categoryMainTextColor, lblFontawesome, categoryPanelColor)
	If materialIconsChr <> "" Then newCat.img = GetBmpFromFontLabel(materialIconsChr, categoryMainTextColor, lblMaterialIcons, categoryPanelColor)
#End If

	If dependentOn.IsInitialized Then
		Dim lst As List
		lst.Initialize2(dependentOn)
		newCat.DependentOn = lst
	End If

	allCategories.Put(newCat.ID, newCat)

	If prefData.ContainsKey(newCat.ID & "iscollapsed") = False Then
		prefData.Put(newCat.ID & "iscollapsed", initiallyCollapsed)
		WritePrefsFile(prefEncryptionKey, False, "", "")
	End If

	Return(ADDCATEGORY_OK)

End Sub

#If B4A
'Adds a preference entry to a category; only required when using the user interface:
'- categoryID: the case-insensitive ID of the category to which this entry belongs
'- ID: a UNIQUE (across all categories!) case-insensitive ID that will be used internally (sometimes combined with categoryID) for the entry; do not use any of the following as the entry won't be added then:
'  - 'IsCollapsed' (case insensitive)
'  - an ID that would cause the value of categoryID.Trim.ToLowerCase &amp; entryID.Trim.ToLowerCase to be the value of one of these private variables defined in the class's Class_Globals sub: prefClassID, prefClassLastUpdated
'- mainText: the entry's title
'- subText: a text to display underneath mainText in the entry panel in the preferences CustomListView
'- longClickHint: the text to display when the user long-clicks the entry
'- inputType: sets the return data type for the entry:
'  - INPUT_TYPE_BOOLEAN: Boolean, True/False; choices: not applicable (use Null)
'  - INPUT_TYPE_EDITTEXT_ANY: String, any text; choices: not applicable (use Null)
'  - INPUT_TYPE_EDITTEXT_PASSWORD: String, any text; choices: not applicable (use Null)
'  - INPUT_TYPE_EDITTEXT_NUMBER: a decimal number; choices: not applicable (use Null)
'  - INPUT_TYPE_EDITTEXT_INTINCL: an Int; choices: Null, or a List containing the minimum and maximum values in elements 0 and 1; both values inclusive
'  - INPUT_TYPE_EDITTEXT_INTEXCL: an Int; choices: Null, or a List containing the minimum and maximum values in elements 0 and 1; both values exclusive
'  - INPUT_TYPE_INPUTLIST: String, one value to be selected with an InputList; choices: a List containing the values to choose from
'  - INPUT_TYPE_INPUTMULTILIST: String, multiple values can be selected with an InputMap; choices: a List containing the values to choose from
'  - INPUT_TYPE_FOLDER: String, a directory that can be selected with a FileDialog; choices: not applicable (use Null)
'  - INPUT_TYPE_FILE: String, a file that can be selected with a FileDialog; choices: a list containing the initial directory- and file names in elements 0 and 1; Null values mean 'don't present that initial value'
'  - INPUT_TYPE_DATE: an Int date that can be selected with a DateDialog; choices: not applicable (use Null)
'  - INPUT_TYPE_TIME: an Int time that can be selected with a TimeDialog; choices: not applicable (use Null)
'  - INPUT_TYPE_DATEANDTIME: a List containing an Int data and Int time in elements 0 and 1 respectively, selected with a DateDialog and a TimeDialog; choices: not applicable (use Null)
'  - INPUT_TYPE_INTENT: does not return data; choices: a list with 1 element being the initialized intent for which StartActivity will be called when the entry is clicked
'- choices: a List that, for the relevant inputTypes (see above), contains the values from which the user can choose
'- defaultValues: is ignored when it is Null, and for entries that are already present in the preferences file; otherwise it can contain the default value(s) for the new entry in element 0; for entry type 'INPUT_TYPE_INPUTMULTILIST', multiple selected values can be specified in multiple elements
'- dependentOn: an empty string, or the entry ID of an entry (from the current entry's category) of type 'INPUT_TYPE_BOOLEAN'; if that checkbox is checked, the new entry will be shown; when it is unchecked, the entry will be hidden
'- fontAwesomeChr: "" or the Chr() value of the FontAwesome character to use at the right of mainText; ignored if showDefaultIconsForEntries = True
'- materialIconsChr: "" or the Chr() value of the MaterialIcons character to use in front of mainText; ignored if showDefaultIconsForEntries = True
'- gotoScreen: "" or the screen ID that was added with AddScreen; only relevant if inputType is INPUT_TYPE_SCREEN
'- Returns: ADDENTRY_OK if all went well, or one of the other ADDENTRY_... values in case an error occurred
Public Sub AddEntry(categoryID As String, ID As String, mainText As String, subText As String, longClickHint As String, inputType As Int, choices As List, defaultValues As List, dependentOn As String, fontAwesomeChr As String, materialIconsChr As String, gotoScreen As String) As Int
	Return AddEntryPrivate(categoryID, ID, mainText, subText, longClickHint, inputType, choices, defaultValues, dependentOn, fontAwesomeChr, materialIconsChr, gotoScreen, False)
#Else If B4J
'Adds a preference entry to a category; only required when using the user interface:
'- categoryID: the case-insensitive ID of the category to which this entry belongs
'- ID: a UNIQUE (across all categories!) case-insensitive ID that will be used internally (sometimes combined with categoryID) for the entry; do not use any of the following as the entry won't be added then:
'  - 'IsCollapsed' (case insensitive)
'  - an ID that would cause the value of categoryID.Trim.ToLowerCase &amp; entryID.Trim.ToLowerCase to be the value of one of these private variables defined in the class's Class_Globals sub: prefClassID, prefClassLastUpdated
'- mainText: the entry's title
'- subText: a text to display underneath mainText in the entry panel in the preferences CustomListView
'- longClickHint: the text to display when the user long-clicks the entry
'- inputType: sets the return data type for the entry:
'  - INPUT_TYPE_BOOLEAN: Boolean, True/False; choices: not applicable (use Null)
'  - INPUT_TYPE_EDITTEXT_ANY: String, any text; choices: not applicable (use Null)
'  - INPUT_TYPE_EDITTEXT_PASSWORD: String, any text; choices: not applicable (use Null)
'  - INPUT_TYPE_EDITTEXT_NUMBER: a decimal number; choices: not applicable (use Null)
'  - INPUT_TYPE_EDITTEXT_INTINCL: an Int; choices: Null, or a List containing the minimum and maximum values in elements 0 and 1; both values inclusive
'  - INPUT_TYPE_EDITTEXT_INTEXCL: an Int; choices: Null, or a List containing the minimum and maximum values in elements 0 and 1; both values exclusive
'  - INPUT_TYPE_INPUTLIST: String, one value to be selected with an InputList; choices: a List containing the values to choose from
'  - INPUT_TYPE_INPUTMULTILIST: String, multiple values can be selected with an InputMap; choices: a List containing the values to choose from
'  - INPUT_TYPE_FOLDER: String, a directory that can be selected with a FileDialog; choices: not applicable (use Null)
'  - INPUT_TYPE_FILE: String, a file that can be selected with a FileDialog; choices: a list containing the initial directory- and file names in elements 0 and 1; Null values mean 'don't present that initial value'
'  - INPUT_TYPE_DATE: an Int date that can be selected with a DateDialog; choices: not applicable (use Null)
'  - INPUT_TYPE_TIME: an Int time that can be selected with a TimeDialog; choices: not applicable (use Null)
'  - INPUT_TYPE_DATEANDTIME: a List containing an Int data and Int time in elements 0 and 1 respectively, selected with a DateDialog and a TimeDialog; choices: not applicable (use Null)
'- choices: a List that, for the relevant inputTypes (see above), contains the values from which the user can choose
'- defaultValues: is ignored when it is Null, and for entries that are already present in the preferences file; otherwise it can contain the default value(s) for the new entry in element 0; for entry type 'INPUT_TYPE_INPUTMULTILIST', multiple selected values can be specified in multiple elements
'- dependentOn: an empty string, or the entry ID of an entry (from the current entry's category) of type 'INPUT_TYPE_BOOLEAN'; if that checkbox is checked, the new entry will be shown; when it is unchecked, the entry will be hidden
'- fontAwesomeChr: "" or the Chr() value of the FontAwesome character to use at the right of mainText; ignored if showDefaultIconsForEntries = True
'- materialIconsChr: "" or the Chr() value of the MaterialIcons character to use in front of mainText; ignored if showDefaultIconsForEntries = True
'- gotoScreen: "" or the screen ID that was added with AddScreen; only relevant if inputType is INPUT_TYPE_SCREEN
'- Returns: ADDENTRY_OK if all went well, or one of the other ADDENTRY_... values in case an error occurred
Public Sub AddEntry(categoryID As String, ID As String, mainText As String, subText As String, longClickHint As String, inputType As Int, choices As List, defaultValues As List, dependentOn As String, fontAwesomeChr As String, materialIconsChr As String, gotoScreen As String) As Int
	Return AddEntryPrivate(categoryID, ID, mainText, subText, longClickHint, inputType, choices, defaultValues, dependentOn, fontAwesomeChr, materialIconsChr, gotoScreen, False)
#End If
End Sub

#If B4A
Private Sub AddEntryPrivate(categoryID As String, ID As String, mainText As String, subText As String, longClickHint As String, inputType As Int, choices As List, defaultValues As List, dependentOn As String, fontAwesomeChr As String, materialIconsChr As String, gotoScreen As String, allowReserved As Boolean) As Int
#Else If B4J
Private Sub AddEntryPrivate(categoryID As String, ID As String, mainText As String, subText As String, longClickHint As String, inputType As Int, choices As List, defaultValues As List, dependentOn As String, fontAwesomeChr As String, materialIconsChr As String, gotoScreen As String, allowReserved As Boolean) As Int
#End If

#If B4A
	Dim pnl As Panel
	Dim mainTextVar As CSBuilder
	Dim subTextVar As CSBuilder
#Else If B4J
	Dim pnl As Pane
	Dim mainTextVar As String
	Dim subTextVar As String
#End If

	Dim newEntPnl As PanelType
	Dim newEnt As EntryType
	Dim compoundID As String

	If (inputType < 0) Then
		If allowReserved = False Then Return ADDENTRY_INVALID_INPUT_TYPE
	Else If inputType > max_input_type Then
		Return ADDENTRY_INVALID_INPUT_TYPE
	End If

	If ID.Trim.ToLowerCase = "iscollapsed" Then Return(ADDENTRY_RESERVED_NAME) ' This would be a reserved name

	If allCategories.ContainsKey(categoryID.Trim.ToLowerCase) = False Then Return(ADDENTRY_CATEGORY_DOES_NOT_EXIST)

	compoundID = categoryID.Trim.ToLowerCase & ID.Trim.ToLowerCase

	Select Case compoundID
		Case prefClassID.ToLowerCase, prefClassLastUpdated.ToLowerCase
			Return(ADDENTRY_RESERVED_NAME) ' These would be reserved names
	End Select

	pnl.Initialize("Entry")

	newEntPnl.Pnl = pnl
	newEntPnl.Color = entryPanelColor
	newEntPnl.BorderWidth = entryPanelBorderWidth
	newEntPnl.BorderRadius = entryPanelBorderRadius
	newEntPnl.BorderColor = entryPanelBorderColor

#If B4A
	If useDefaultIcons Then
		Select Case inputType
			Case INPUT_TYPE_BACKUP
				materialIconsChr = Chr(0xE161)
			Case INPUT_TYPE_RESTORE
				materialIconsChr = Chr(0xE929)
			Case INPUT_TYPE_RESET
				materialIconsChr = Chr(0xE8BA)
			Case INPUT_TYPE_DATE, INPUT_TYPE_DATEANDTIME
				fontAwesomeChr = Chr(0xF073)
			Case INPUT_TYPE_TIME
				materialIconsChr = Chr(0xE192)
			Case INPUT_TYPE_FOLDER
				fontAwesomeChr = Chr(0xF115)
			Case INPUT_TYPE_FILE
				fontAwesomeChr = Chr(0xF016)
			Case INPUT_TYPE_BOOLEAN
				' Do nothing; this is handled in Sub AssembleprefXCLVentries
			Case INPUT_TYPE_INPUTLIST, INPUT_TYPE_INPUTMULTILIST
				fontAwesomeChr = Chr(0xF03A)
			Case INPUT_TYPE_EDITTEXT_INTINCL, INPUT_TYPE_EDITTEXT_INTEXCL
				materialIconsChr = Chr(0xE242)
			Case INPUT_TYPE_EDITTEXT_NUMBER
				fontAwesomeChr = Chr(0xF292)
			Case INPUT_TYPE_SCREEN
				fontAwesomeChr = Chr(0xF2D2)
			Case Else
				materialIconsChr = Chr(0xE165)
		End Select
	End If
	If fontAwesomeChr <> "" Then
		mainTextVar.Initialize.Color(entryMainTextColor).size(entryMainTextSize).Typeface(Typeface.FONTAWESOME).Append(fontAwesomeChr).Pop.Append(" " & mainText).PopAll
	Else If materialIconsChr <> "" Then
		mainTextVar.Initialize.Color(entryMainTextColor).size(entryMainTextSize).Typeface(Typeface.MATERIALICONS).Append(materialIconsChr).Pop.Append(" " & mainText).PopAll
	Else
		mainTextVar.Initialize.Color(entryMainTextColor).size(entryMainTextSize).Append(mainText).PopAll
	End If
	If subText = "" Then
		subTextVar.Initialize
	Else If fontAwesomeChr <> "" Then
		subTextVar.Initialize.Size(entryMainTextSize).Color(entryPanelColor).Typeface(Typeface.FONTAWESOME).Append(fontAwesomeChr).Pop.Append(" ").PopAll.Size(entrySubTextSize).Color(entrySubTextColor).Append(subText).PopAll
	Else If materialIconsChr <> "" Then
		subTextVar.Initialize.Size(entryMainTextSize).Color(entryPanelColor).Typeface(Typeface.MATERIALICONS).Append(materialIconsChr).Pop.Append(" ").PopAll.Size(entrySubTextSize).Color(entrySubTextColor).Append(subText).PopAll
	Else
		subTextVar.Initialize.Color(entrySubTextColor).size(entrySubTextSize).Append(subText).PopAll
	End If
	newEnt.fontAweSomeChar = fontAwesomeChr
	newEnt.materialIconsChar = materialIconsChr
#Else If B4J
	mainTextVar = mainText
	subTextVar = subText
#End If

	newEnt.CategoryID = categoryID.Trim.ToLowerCase
	newEnt.EntryID = ID.Trim.ToLowerCase
	newEnt.CompoundID = compoundID
	newEnt.pnl = newEntPnl
	newEnt.MainText = mainTextVar
	newEnt.SubText = subTextVar
	newEnt.LongClickHint = longClickHint.Trim
	newEnt.EntryInputType = inputType
	newEnt.pnlColor = entryPanelColor
	newEnt.gotoScreen = gotoScreen.Trim.ToLowerCase

#If B4J
	If useDefaultIcons = False Then
		If fontAwesomeChr <> "" Then newEnt.img = GetBmpFromFontLabel(fontAwesomeChr, entryIconsColor, lblFontawesome, entryPanelColor)
		If materialIconsChr <> "" Then newEnt.img = GetBmpFromFontLabel(materialIconsChr, entryIconsColor, lblMaterialIcons, entryPanelColor)
	Else
		Select Case inputType
			Case INPUT_TYPE_BACKUP
				newEnt.img = GetBmpFromFontLabel(Chr(0xE161), entryIconsColor, lblMaterialIcons, entryPanelColor)
			Case INPUT_TYPE_RESTORE
				newEnt.img = GetBmpFromFontLabel(Chr(0xE929), entryIconsColor, lblMaterialIcons, entryPanelColor)
			Case INPUT_TYPE_RESET
				newEnt.img = GetBmpFromFontLabel(Chr(0xE8BA), entryIconsColor, lblMaterialIcons, entryPanelColor)
			Case INPUT_TYPE_DATE, INPUT_TYPE_DATEANDTIME
				newEnt.img = GetBmpFromFontLabel(Chr(0xF073), entryIconsColor, lblFontawesome, entryPanelColor)
			Case INPUT_TYPE_TIME
				newEnt.img = GetBmpFromFontLabel(Chr(0xE192), entryIconsColor, lblMaterialIcons, entryPanelColor)
			Case INPUT_TYPE_FOLDER
				newEnt.img = GetBmpFromFontLabel(Chr(0xF115), entryIconsColor, lblFontawesome, entryPanelColor)
			Case INPUT_TYPE_FILE
				newEnt.img = GetBmpFromFontLabel(Chr(0xF016), entryIconsColor, lblFontawesome, entryPanelColor)
			Case INPUT_TYPE_BOOLEAN
				' Do nothing; this is handled in Sub AssembleprefXCLVentries
			Case INPUT_TYPE_INPUTLIST, INPUT_TYPE_INPUTMULTILIST
				newEnt.img = GetBmpFromFontLabel(Chr(0xF03A), entryIconsColor, lblFontawesome, entryPanelColor)
			Case INPUT_TYPE_EDITTEXT_INTINCL, INPUT_TYPE_EDITTEXT_INTEXCL
				newEnt.img = GetBmpFromFontLabel(Chr(0xE242), entryIconsColor, lblMaterialIcons, entryPanelColor)
			Case INPUT_TYPE_EDITTEXT_NUMBER
				newEnt.img = GetBmpFromFontLabel(Chr(0xF292), entryIconsColor, lblFontawesome, entryPanelColor)
			Case INPUT_TYPE_SCREEN
				newEnt.img = GetBmpFromFontLabel(Chr(0xF2D2), entryIconsColor, lblFontawesome, entryPanelColor)
			Case Else
				newEnt.img = GetBmpFromFontLabel(Chr(0xE165), entryIconsColor, lblMaterialIcons, entryPanelColor)
		End Select
	End If
#End If

	If dependentOn.Trim = "" Then
		newEnt.DependentOn = ""
	Else
		newEnt.DependentOn = categoryID.Trim.ToLowerCase & dependentOn.Trim.ToLowerCase
	End If

	If choices.IsInitialized Then
		Dim lst As List
		lst.Initialize2(choices)
		newEnt.EntryChoices = lst
	End If

	If defaultValues.IsInitialized Then
		Dim lst As List
		lst.Initialize2(defaultValues)
		newEnt.defaultValues = lst
	End If

	allEntries.put(compoundID, newEnt)

	If defaultValues.IsInitialized = False Then Return(ADDENTRY_OK) ' No default values specified
	If defaultValues = Null Then Return(ADDENTRY_OK) ' No default values specified
	If prefData.Get(newEnt.EntryID) <> Null Then Return(ADDENTRY_OK) ' Value already exists; ignore default values

	Select Case newEnt.EntryInputType
		Case INPUT_TYPE_INPUTMULTILIST
			Dim currentSelections As List
			currentSelections.Initialize2(defaultValues)
			prefData.Put(newEnt.EntryID, currentSelections)
		Case Else
			prefData.Put(newEnt.EntryID, defaultValues.Get(0))
	End Select

	Return(ADDENTRY_OK)

End Sub

#End Region ' Setup

#Region UserInteraction

Private Sub Category_Click

#If B4A
	Dim send As View = Sender
#Else If B4J
	Dim send As Node = Sender
#End If
	Dim ett As EventAndTagType = send.Tag
	Dim theCat As CategoryType = ett.tag
	Category_Clicked(theCat, ett.xclvItemIndex)

End Sub

Private Sub Category_Clicked(theCat As CategoryType, idx As Int)

	' Keep the collapsed state in the file so that is it persistent between runs
	prefData.Put(theCat.ID & "iscollapsed", Not(CategoryIsCollapsed(theCat)))
	WritePrefsFile(prefEncryptionKey, False, "", "")
	PushToJumpToStack(idx)
	AssembleprefXCLV

End Sub

Private Sub CategoryIsCollapsed(theCat As CategoryType) As Boolean

	Return(ConvertToBoolean(prefData.Get(theCat.ID & "iscollapsed")))

End Sub

Private Sub Category_LongClick

#If B4A
	Dim send As View = Sender
#Else If B4J
	Dim send As Node = Sender
#End If

	Dim ett As EventAndTagType = send.Tag
	Dim theCat As CategoryType = ett.tag

	If theCat.LongClickHint <> "" Then ShowToast(theCat.LongClickHint, True)

End Sub

Private Sub Entry_Click

#If B4A
	Dim send As View = Sender
#Else If B4J
	Dim send As Node = Sender
#End If
	Dim ett As EventAndTagType = send.Tag
	Dim theEntry As EntryType = ett.Tag
	Entry_Clicked(theEntry, ett.xclvItemIndex)

End Sub

Private Sub Entry_Clicked(theEntry As EntryType, idx As Int)

	Dim dataChanged As Boolean = False
	Dim gotoScreen As Boolean = False

	Select Case theEntry.EntryInputType
		Case INPUT_TYPE_DATE
			dataChanged = Date_ValueChanged(theEntry)
		Case INPUT_TYPE_TIME
			dataChanged = Time_ValueChanged(theEntry)
		Case INPUT_TYPE_DATEANDTIME
			dataChanged = DateTime_ValueChanged(theEntry)
		Case INPUT_TYPE_BOOLEAN
			dataChanged = Checkbox_ValueChanged(theEntry)
		Case INPUT_TYPE_FOLDER
			dataChanged = Folder_ValueChanged(theEntry)
		Case INPUT_TYPE_FILE
			dataChanged = File_ValueChanged(theEntry)
		Case INPUT_TYPE_INPUTLIST
			dataChanged = Inputlist_ValueChanged(theEntry)
		Case INPUT_TYPE_INPUTMULTILIST
			dataChanged = Inputmultilist_ValueChanged(theEntry)
		Case INPUT_TYPE_EDITTEXT_ANY
			dataChanged = EdittextAny_ValueChanged(theEntry, False)
		Case INPUT_TYPE_EDITTEXT_PASSWORD
			dataChanged = EdittextAny_ValueChanged(theEntry, True)
		Case INPUT_TYPE_EDITTEXT_NUMBER
			dataChanged = EdittextNumber_ValueChanged(theEntry)
		Case INPUT_TYPE_EDITTEXT_INTINCL, INPUT_TYPE_EDITTEXT_INTEXCL
			dataChanged = EdittextInt_ValueChanged(theEntry)
#If B4A
		Case INPUT_TYPE_INTENT
			dataChanged = Intent_ValueChanged(theEntry)
#End If
		Case INPUT_TYPE_BACKUP
			Backup
		Case INPUT_TYPE_RESTORE
			Restore
		Case INPUT_TYPE_RESET
			Reset
		Case INPUT_TYPE_SCREEN
			gotoScreen = True
			screenStack.InsertAt(0, theEntry.gotoScreen)
		Case INPUT_TYPE_BACK
			gotoScreen = True
			screenStack.RemoveAt(0)
		Case Else
			Return
	End Select

	If dataChanged Then
		WritePrefsFile(prefEncryptionKey, False, "", "")
		PushToJumpToStack(idx)
		AssembleprefXCLV
	Else If gotoScreen Then
		If theEntry.EntryInputType <> INPUT_TYPE_BACK Then
			PushToJumpToStack(idx)
			PushToJumpToStack(0)
		End If
		AssembleprefXCLV
	End If

#If B4A
	ime1.HideKeyboard
#End If

End Sub

#If B4A
Private Sub Intent_ValueChanged(theEntry As EntryType) As Boolean

	StartActivity(theEntry.EntryChoices.Get(0))
	Return(False)

End Sub
#End If

Private Sub Date_ValueChanged(theEntry As EntryType) As Boolean

	Dim curVal As String = ""

	If prefData.ContainsKey(theEntry.EntryID) Then curVal = prefData.Get(theEntry.EntryID)

#If B4A
	Dim dd As DateDialog

	Try
		dd.Year = DateTime.GetYear(curVal)
		dd.Month = DateTime.GetMonth(curVal)
		dd.DayOfMonth = DateTime.GetDayOfMonth(curVal)
	Catch
		dd.Year = DateTime.GetYear(DateTime.Now)
		dd.Month = DateTime.GetMonth(DateTime.Now)
		dd.DayOfMonth = DateTime.GetDayOfMonth(DateTime.Now)
	End Try

	If dd.Show(theEntry.SubText, theEntry.MainText, prefOKtext, prefCancelText, "", Null) <> DialogResponse.POSITIVE Then Return(False)

	prefData.Put(theEntry.EntryID, dd.DateTicks)

	Return(True)
#Else If B4J
	Dim dateDialog As DialogsX
	Dim saveDateFormat As String = DateTime.DateFormat
	Dim selectedDate As String = ""

	dateDialog.Initialize
	dateDialog.OKButtonText = prefOKtext
	dateDialog.CancelButtonText = prefCancelText
	dateDialog.SetParentWindow(callingForm)

	DateTime.DateFormat = "yyyy-MM-dd"
	If curVal = "" Then
		selectedDate = dateDialog.DatePickerDialog(theEntry.MainText, theEntry.SubText, "", "")
	Else
		selectedDate = dateDialog.DatePickerDialog(theEntry.MainText, theEntry.SubText, "", DateTime.Date(curVal))
	End If

	If selectedDate = "" Then
		DateTime.DateFormat = saveDateFormat
		Return(False)
	End If

	prefData.Put(theEntry.EntryID, DateTime.DateParse(selectedDate))
	DateTime.DateFormat = saveDateFormat
	Return(True)
#End If

End Sub

Private Sub Time_ValueChanged(theEntry As EntryType) As Boolean

	Dim curVal As String = ""
	Dim theHour As Int
	Dim theMinute As Int

	If prefData.ContainsKey(theEntry.EntryID) Then curVal = prefData.Get(theEntry.EntryID)

	Try
		theHour = DateTime.GetHour(curVal)
		theMinute = DateTime.GetMinute(curVal)
	Catch
		theHour = DateTime.GetHour(DateTime.Now)
		theMinute = DateTime.GetMinute(DateTime.Now)
	End Try

#If B4A
	Dim Td As TimeDialog

	If prefData.ContainsKey(theEntry.EntryID) Then curVal = prefData.Get(theEntry.EntryID)

	Td.Hour = theHour
	Td.Minute = theMinute

	If Td.Show(theEntry.SubText, theEntry.MainText, prefOKtext, prefCancelText, "", Null) <> DialogResponse.POSITIVE Then Return(False)

	prefData.Put(theEntry.EntryID, Td.TimeTicks)
	Return(True)
#Else If B4J
	Dim timeDialog As DialogsX
	Dim selectedTime As String = ""

	timeDialog.Initialize
	timeDialog.OKButtonText = prefOKtext
	timeDialog.CancelButtonText = prefCancelText
	timeDialog.SetParentWindow(callingForm)

	selectedTime = timeDialog.TimePicker24Dialog(theEntry.MainText, theEntry.SubText, theHour, theMinute, False)
	If selectedTime = "" Then Return(False)

	prefData.Put(theEntry.EntryID, DateTime.TimeParse(selectedTime & ":00"))
	Return(True)
#End If

End Sub

Private Sub DateTime_ValueChanged(theEntry As EntryType) As Boolean

	Dim valList As List

	valList.Initialize

	Try
		If prefData.ContainsKey(theEntry.EntryID) Then valList.AddAll(prefData.Get(theEntry.EntryID))
	Catch
		Log("") ' No need for action here
	End Try

	If valList.Size <> 2 Then
		valList.Clear
		valList.Add(DateTime.Now)
		valList.Add(DateTime.Now)
	End If

#If B4A
	Dim Dd As DateDialog
	Dim Td As TimeDialog

	Dd.Year = DateTime.GetYear(valList.Get(0))
	Dd.Month = DateTime.GetMonth(valList.Get(0))
	Dd.DayOfMonth = DateTime.GetDayOfMonth(valList.Get(0))

	If Dd.Show(theEntry.SubText, theEntry.MainText, prefOKtext, prefCancelText, "", Null) <> DialogResponse.POSITIVE Then Return(False)

	Td.Hour = DateTime.GetHour(valList.Get(1))
	Td.Minute = DateTime.GetMinute(valList.Get(1))

	If Td.Show(theEntry.SubText, theEntry.MainText, prefOKtext, prefCancelText, "", Null) <> DialogResponse.POSITIVE Then Return(False)

	valList.Clear
	valList.Add(Dd.DateTicks)
	valList.Add(Td.TimeTicks)
	prefData.Put(theEntry.EntryID, valList)

	Return(True)
#Else If B4J
	Dim dateTimeDialog As DialogsX
	Dim saveDateFormat As String = DateTime.DateFormat
	Dim selectedDate As String = ""
	Dim selectedTime As String = ""

	dateTimeDialog.Initialize
	dateTimeDialog.OKButtonText = prefOKtext
	dateTimeDialog.CancelButtonText = prefCancelText
	dateTimeDialog.SetParentWindow(callingForm)

	DateTime.DateFormat = "yyyy-MM-dd"
	If valList.Get(0) = "" Then
		selectedDate = dateTimeDialog.DatePickerDialog(theEntry.MainText, theEntry.SubText, "", "")
	Else
		selectedDate = dateTimeDialog.DatePickerDialog(theEntry.MainText, theEntry.SubText, "", DateTime.Date(valList.Get(0)))
	End If

	If selectedDate = "" Then
		DateTime.DateFormat = saveDateFormat
		Return(False)
	End If

	If valList.Get(1) = "" Then
		selectedTime = dateTimeDialog.TimePicker24Dialog(theEntry.MainText, theEntry.SubText, DateTime.GetHour(DateTime.Now), DateTime.GetMinute(DateTime.Now), False)
	Else
		selectedTime = dateTimeDialog.TimePicker24Dialog(theEntry.MainText, theEntry.SubText, DateTime.GetHour(valList.Get(1)), DateTime.GetMinute(valList.Get(1)), False)
	End If

	If selectedTime = "" Then
		DateTime.DateFormat = saveDateFormat
		Return(False)
	End If

	valList.Clear
	valList.Add(DateTime.DateParse(selectedDate))
	valList.Add(DateTime.TimeParse(selectedTime & ":00"))
	prefData.Put(theEntry.EntryID, valList)
	DateTime.DateFormat = saveDateFormat
	Return(True)
#End If

End Sub

Private Sub Folder_ValueChanged(theEntry As EntryType) As Boolean

	Dim fullPath As String = ""

#If B4A
	Dim fd As FileDialog
	fd.FastScroll = True
	fd.ShowOnlyFolders = True
	fd.FilePath = rp.GetSafeDirDefaultExternal("")
	If prefData.ContainsKey(theEntry.EntryID) Then
		fullPath = prefData.Get(theEntry.EntryID)
		If fullPath <> "" Then
			If File.Exists(fullPath, "") Then fd.FilePath = fullPath
		End If
	End If
	If fd.Show(theEntry.MainText, prefOKtext, prefCancelText, "", Null) <> DialogResponse.POSITIVE Then Return(False)
	fullPath = File.Combine(fd.FilePath, fd.ChosenName)
#Else If B4J
	Dim dc As DirectoryChooser
	dc.Initialize
	dc.Title = theEntry.MainText
	If prefData.ContainsKey(theEntry.EntryID) Then dc.InitialDirectory = prefData.Get(theEntry.EntryID)
	fullPath = dc.Show(callingForm)
#End If

	Select Case fullPath
		Case "", Null
			Return(False)
	End Select

	If File.Exists(fullPath, "") = False Then File.MakeDir(fullPath, "")

	prefData.Put(theEntry.EntryID, fullPath)
	Return(True)

End Sub

Private Sub File_ValueChanged(theEntry As EntryType) As Boolean

	Dim fname As String
	Dim fldr As String = ""

#If B4A
	Dim fd As FileDialog
	fd.FastScroll = True
	fd.FilePath = rp.GetSafeDirDefaultExternal("")
	If prefData.ContainsKey(theEntry.EntryID) Then
		Try
			fd.ChosenName = prefData.Get(theEntry.EntryID)
		Catch
			Log("File_ValueChanged: " & LastException)
		End Try
	End If

	If fd.Show(theEntry.MainText, prefOKtext, prefCancelText, "", Null) <> DialogResponse.POSITIVE Then Return(False)
	fname = fd.ChosenName
	fldr = fd.FilePath
#Else
	Dim fc As FileChooser
	fc.Initialize
	fc.Title = theEntry.MainText
	If prefData.ContainsKey(theEntry.EntryID) Then fc.InitialFileName = prefData.Get(theEntry.EntryID)
	fname = fc.ShowOpen(callingForm)
#End If

	If fname = "" Then Return(False)
	If fname = Null Then Return(False)

	prefData.Put(theEntry.EntryID, File.Combine(fldr, fname))
	Return(True)

End Sub

Private Sub Checkbox_ValueChanged(theEntry As EntryType) As Boolean

	Dim curVal As Boolean = False

	Try
		curVal = prefData.GetDefault(theEntry.EntryID, False)
		prefData.Put(theEntry.EntryID, Not(curVal))
	Catch
		prefData.Put(theEntry.EntryID, False)
	End Try

	Return(True)

End Sub

Private Sub EdittextAny_ValueChanged(theEntry As EntryType, passwordMode As Boolean) As Boolean

	Dim curVal As String = ""
	If prefData.ContainsKey(theEntry.EntryID) Then curVal = prefData.Get(theEntry.EntryID)

#If B4A
	Dim showPw As Boolean = False
	Dim reslt As Int

	Do While True
		Dim inpDlg As InputDialog
		inpDlg.Input = curVal
		inpDlg.Hint = theEntry.SubText
		If passwordMode Then
			If showPw Then
				reslt = inpDlg.Show(theEntry.MainText, prefTitleText, prefOKtext, prefCancelText, prefHideText, Null)
			Else
				inpDlg.PasswordMode = passwordMode
				reslt = inpDlg.Show(theEntry.MainText, prefTitleText, prefOKtext, prefCancelText, prefShowText, Null)
			End If
		Else
			reslt = inpDlg.Show(theEntry.MainText, prefTitleText, prefOKtext, prefCancelText, "", Null)
		End If
		Select Case reslt
			Case DialogResponse.POSITIVE
				prefData.Put(theEntry.EntryID, inpDlg.Input.trim)
				ime1.HideKeyboard
				Return(True)
			Case DialogResponse.CANCEL
				ime1.HideKeyboard
				Return(False)
			Case Else
				showPw = Not(showPw)
		End Select
	Loop
	Return(False)
#Else If B4J
	Dim inpDlg As DialogsX
	Dim inpDlgResult As String

	If passwordMode Then
		' Based on https://www.b4x.com/android/forum/threads/jrldialogsx-dialogs-library.52416/page-3#post-506883
		inpDlg.Initialize
		' Create the list of fieldmaps
		Dim fieldlist As List
		fieldlist.Initialize
		Dim m As Map = CreateMap("id":"password", "label":theEntry.SubText, "value":curVal, "type":"P")
		fieldlist.Add(m)
		Dim m As Map = CreateMap("id":"password2", "label":prefRepeatPwText, "value":"", "type":"P")
		fieldlist.Add(m)
		' Open the dialog
		Dim resultmap As Map = inpDlg.SimpleFormDialog(prefTitleText, theEntry.MainText, fieldlist)
		' The resultmap holds for each field the pair id:value
		If resultmap.IsInitialized Then
			inpDlgResult = resultmap.Get("password")
			If inpDlgResult <> resultmap.Get("password2") Then
				ShowMsgBox(prefInvalidPwText, prefTitleText)
				Return False
			End If
		Else
			Return(False)
		End If
	Else
		inpDlg.Initialize
		inpDlg.OKButtonText = prefOKtext
		inpDlg.CancelButtonText = prefCancelText
		inpDlg.SetParentWindow(callingForm)
		inpDlgResult = inpDlg.TextInputDialog3(prefTitleText, theEntry.MainText, theEntry.SubText, curVal)
		If inpDlgResult.EqualsIgnoreCase(Null) Then Return(False)
	End If

	prefData.Put(theEntry.EntryID, inpDlgResult.trim)
	Return(True)
#End If

End Sub

Private Sub EdittextNumber_ValueChanged(theEntry As EntryType) As Boolean

	Dim curVal As String = ""
	If prefData.ContainsKey(theEntry.EntryID) Then curVal = prefData.Get(theEntry.EntryID)

#If B4A
	Dim inpDlg As InputDialog
	inpDlg.Input = curVal
	inpDlg.Hint = theEntry.SubText
	inpDlg.InputType = inpDlg.INPUT_TYPE_DECIMAL_NUMBERS
	If inpDlg.Show(theEntry.MainText, prefTitleText, prefOKtext, prefCancelText, "", Null) <> DialogResponse.POSITIVE Then
		ime1.HideKeyboard
		Return(False)
	End If
	prefData.Put(theEntry.EntryID, inpDlg.Input.trim)
	ime1.HideKeyboard
	Return(True)
#Else If B4J
	Dim inpDlg As DialogsX
	Dim inpDlgResult As String
	inpDlg.Initialize
	inpDlg.OKButtonText = prefOKtext
	inpDlg.CancelButtonText = prefCancelText
	inpDlg.SetParentWindow(callingForm)
	Do While True
		inpDlgResult = inpDlg.TextInputDialog3(prefTitleText, theEntry.MainText, theEntry.SubText, curVal).Trim
		If inpDlgResult.EqualsIgnoreCase(Null) Then Return(False)
		If (IsSignedDecimal(inpDlgResult, False, theDecimalSeparator)) Or (IsUnsignedDecimal(inpDlgResult, False, theDecimalSeparator)) Then
			prefData.Put(theEntry.EntryID, inpDlgResult.trim)
			Return(True)
		End If
	Loop
	Return(False)
#End If

End Sub

Private Sub EdittextInt_ValueChanged(theEntry As EntryType) As Boolean

	Dim anyWillDo As Boolean = False
	Dim valList As List
	Dim i As Int
	Dim iFrom As Int
	Dim iTo As Int
	Dim inputListData As List

	' Will any value do?
	If theEntry.EntryChoices = Null Then
		anyWillDo = True
	Else
		valList.Initialize
		valList.AddAll(theEntry.EntryChoices)
		If valList.Size <> 2 Then anyWillDo = True
	End If

	If anyWillDo Then
		Dim curVal As Int = 0
		If prefData.ContainsKey(theEntry.EntryID) Then curVal = prefData.Get(theEntry.EntryID)
#If B4A
		Dim inpDlg As InputDialog
		inpDlg.Input = curVal
		inpDlg.Hint = theEntry.SubText
		inpDlg.InputType = inpDlg.INPUT_TYPE_NUMBERS
		If inpDlg.Show(theEntry.MainText, prefTitleText, prefOKtext, prefCancelText, "", Null) <> DialogResponse.POSITIVE Then
			ime1.HideKeyboard
			Return(False)
		End If
		prefData.Put(theEntry.EntryID, inpDlg.Input.trim)
		ime1.HideKeyboard
		Return(True)
#Else If B4J
		Dim inpDlg As DialogsX
		Dim inpDlgResult As String
		inpDlg.Initialize
		inpDlg.OKButtonText = prefOKtext
		inpDlg.CancelButtonText = prefCancelText
		inpDlg.SetParentWindow(callingForm)
		Do While True
			inpDlgResult = inpDlg.TextInputDialog3(prefTitleText, theEntry.MainText, theEntry.SubText, curVal).Trim
			If inpDlgResult.EqualsIgnoreCase(Null) Then Return(False)
			If (IsSignedInteger(inpDlgResult, False)) Or (IsUnsignedInteger(inpDlgResult, False)) Then
				prefData.Put(theEntry.EntryID, inpDlgResult.trim)
				Return(True)
			End If
		Loop
		Return(False)
#End If
	Else
		inputListData.Initialize
		If theEntry.EntryInputType = INPUT_TYPE_EDITTEXT_INTINCL Then
			iFrom = valList.Get(0)
			iTo = valList.Get(1)
		Else
			iFrom = valList.Get(0) + 1
			iTo = valList.Get(1) - 1
		End If
		For i=iFrom To iTo
			inputListData.Add(i)
		Next
		Return(HandleInputList(inputListData, theEntry))
	End If

End Sub

Private Sub Inputlist_ValueChanged(theEntry As EntryType) As Boolean

	Dim inputListData As List
	inputListData.Initialize2(theEntry.EntryChoices)
	Return(HandleInputList(inputListData, theEntry))

End Sub

Private Sub HandleInputList(Xvalues As List, theEntry As EntryType) As Boolean

	Dim intRes As Int = -1
	Dim checkedItem As Int = -1
	Dim values As List

	values.Initialize
	values.AddAll(Xvalues)
	values.add(prefCancelText)

	If prefData.ContainsKey(theEntry.EntryID) Then
		If prefData.get(theEntry.EntryID) <> Null Then checkedItem = values.IndexOf(prefData.get(theEntry.EntryID))
	End If

#If B4A
	intRes = InputList(values, theEntry.MainText, checkedItem)
#Else If B4J
	intRes = fx.InputList(callingForm, values, theEntry.SubText, theEntry.MainText, checkedItem)
#End if
	Select Case intRes
#If B4A
		Case DialogResponse.CANCEL
			Return(False)
#End if
		Case (values.Size - 1)
			Return(False)
		Case Else
			Try
				prefData.Put(theEntry.EntryID, values.Get(intRes))
				Return(True)
			Catch
				Return False
			End Try
	End Select

End Sub

Private Sub Inputmultilist_ValueChanged(theEntry As EntryType) As Boolean

	Dim inputListData As List : inputListData.Initialize2(theEntry.EntryChoices)
	Dim currentSelections As List
	
	currentSelections.Initialize

	Try
		currentSelections.AddAll(prefData.get(theEntry.EntryID))
	Catch
		Log("") ' This can occur if no selection has been made yet
	End Try

#If B4A
	Dim options As Map
	Dim i As Int
	Dim s As String
	Dim isSelected As Boolean

	options.Initialize
	For i = 0 To (inputListData.Size - 1)
		isSelected = False
		s = inputListData.Get(i)
		If currentSelections.IndexOf(s) > -1 Then isSelected = True
		options.Put(s, isSelected)
	Next
	InputMap(options, theEntry.MainText)

	currentSelections.Clear
	For i = 0 To (options.Size - 1)
		If options.GetValueAt(i) Then currentSelections.Add(options.GetKeyAt(i))
	Next

	prefData.Put(theEntry.EntryID, currentSelections)
	Return(True)
#Else If B4J
	Dim selectedMap As Map
	Dim dlgInputMultiList As DialogsX
	dlgInputMultiList.Initialize

	selectedMap = dlgInputMultiList.ListDialog2(theEntry.MainText, theEntry.SubText, inputListData, 0)
	If selectedMap.IsInitialized = False Then Return(False)

	currentSelections.Clear
	For i = 0 To (selectedMap.Size - 1)
		currentSelections.Add(selectedMap.GetValueAt(i))
	Next

	prefData.Put(theEntry.EntryID, currentSelections)
	Return(True)
#End If

End Sub

Private Sub Entry_LongClick

#If B4A
	Dim send As View = Sender
#Else If B4J
	Dim send As Node = Sender
#End If

	Dim theEntry As EntryType = send.Tag

	If theEntry.LongClickHint <> "" Then ShowToast(theEntry.LongClickHint, True)

End Sub

Private Sub GetEncryptionKey(deflt As String) As String

#If B4A
	Dim inpDlg As InputDialog
	inpDlg.Input = deflt
	If deflt = "" Then inpDlg.PasswordMode = True ' If default value *is* provided, it's a new 4-digit PIN that's being suggested, so let's show it
	If inpDlg.Show(prefEncryptionKeyPrompt, prefTitleText, prefOKtext, prefCancelText, "", Null) <> DialogResponse.POSITIVE Then
		ime1.HideKeyboard
		Return("")
	End If
	ime1.HideKeyboard
	Return(inpDlg.Input.trim)
#Else If B4J
	Dim inpDlg As DialogsX
	inpDlg.Initialize
	inpDlg.OKButtonText = prefOKtext
	inpDlg.CancelButtonText = prefCancelText
	inpDlg.SetParentWindow(callingForm)
	Return inpDlg.TextInputDialog(prefTitleText, prefEncryptionKeyPrompt,  "", deflt).Trim
#End If

End Sub

#End Region ' UserInteraction

#Region GetSet

'Read-only: True if the CustomListView is visible, False if not.
Public Sub Visible() As Boolean

	If prefXCLV.IsInitialized Then
		Return(prefXCLV.AsView.Visible)
	Else
		Return(False)
	End If

End Sub

'Returns all preferences as a Map.
Public Sub GetAll() As Map

	Return prefData

End Sub

'Returns the Boolean value of the preference identified by entryID. Returns False if the value is not found or if it is not a Boolean.
Public Sub GetBoolean(entryID As String) As Boolean

	If prefData.ContainsKey(entryID.Trim.ToLowerCase) Then
		Try
			Dim s As String
			s = prefData.Get(entryID.Trim.ToLowerCase)
			Return(ConvertToBoolean(s))
		Catch
			Return(False)
		End Try
	Else
		Return(False)
	End If

End Sub

'Returns the List containing the values of the 'INPUT_TYPE_INPUTMULTILIST' type preference identified by entryID. Returns Null if no values are found.
Public Sub GetList(entryID As String) As List

	If prefData.ContainsKey(entryID.Trim.ToLowerCase) Then
		Try
			Dim lst As List : lst.Initialize2(prefData.Get(entryID.Trim.ToLowerCase))
			Return(lst)
		Catch
			Return(Null)
		End Try
	Else
		Return(Null)
	End If

End Sub

'Returns the String value of the preference identified by entryID. Returns an empty string if the value is not found.
Public Sub GetString(entryID As String) As String

	If prefData.ContainsKey(entryID.Trim.ToLowerCase) Then
		Try
			Return(prefData.Get(entryID.Trim.ToLowerCase))	
		Catch
			Return("")
		End Try
	Else
		Return("")
	End If

End Sub

'Writes the preferences to the file, e.g. after importing them and replacing prefData
Public Sub Save

	WritePrefsFile(prefEncryptionKey, False, "", "")

End Sub

'Sets the Boolean value of the preference identified by entryID.
Public Sub SetBoolean(entryID As String, val As Boolean)

	prefData.Put(entryID.Trim.ToLowerCase, val)
	WritePrefsFile(prefEncryptionKey, False, "", "")

End Sub

'Sets the values for the 'INPUT_TYPE_INPUTMULTILIST' type preference identified by entryID.
Public Sub SetList(entryID As String, val As List)

	Dim lst As List : lst.Initialize2(val)
	prefData.Put(entryID.Trim.ToLowerCase, lst)
	WritePrefsFile(prefEncryptionKey, False, "", "")

End Sub

'Sets the String value of the preference identified by entryID.
Public Sub SetString(entryID As String, val As String)

	prefData.Put(entryID.Trim.ToLowerCase, val)
	WritePrefsFile(prefEncryptionKey, False, "", "")

End Sub

'Prompts the user for a file path and an (optional) encryption key, and creates a backup of the preferences in the selected file. This file is overwritten if it already exists.
'Does nothing if the user interface was not enabled in the Initialize call.
Public Sub Backup

	If prefXCLV.IsInitialized = False Then Return

	Dim fdFile As String
	Dim fdFolder As String = ""

#If B4A
	Dim fd As FileDialog
	fd.FastScroll = True
	fd.FilePath = rp.GetSafeDirDefaultExternal("")
	If fd.Show(prefBackupText, prefOKtext, prefCancelText, "", Null) <> DialogResponse.POSITIVE Then Return
	fdFile = fd.ChosenName
	fdFolder = fd.FilePath
	If fdFile = "" Then Return
#Else If B4J
	Dim fc As FileChooser
	fc.Initialize
	fc.InitialDirectory = dirData
	fc.Title = prefBackupText
	fdFile = fc.ShowSave(callingForm)
	If fdFile = "" Then Return
#End If

	Try
		If File.Exists(fdFolder, fdFile) Then File.Delete(fdFolder, fdFile)
		WritePrefsFile(GetEncryptionKey(NumberFormat(RandomNumber(0, 9999, 0, True), 4, 0).Replace(",", "")), True, fdFolder, fdFile)
	Catch
		ShowMsgBox(fdFolder & CRLF & fdFile & CRLF & LastException, prefTitleText)
		Return
	End Try

	ShowToast(prefCompletedText & CRLF & File.Combine(fdFolder, fdFile), True)

End Sub

Private Sub RandomNumber(Lowest As Double, Highest As Double, DecimalPlaces As Int, PreventZero As Boolean) As Double

	' Code from https://www.b4x.com/android/forum/threads/randomnumber-the-most-powerful-number-generator-b4a-b4j-b4i.58456/
	' Returns a random double between Lowest (inclusive) and Highest (exclusive), with n decimal places.

    Lowest = Round(Lowest)
    Highest = Round(Highest)
    Dim Decimal As Double
    If DecimalPlaces > 0 Then Decimal = (Rnd(0, Power(10, DecimalPlaces))) / Power(10, DecimalPlaces)
    If Lowest = Highest Then
        Return Lowest
    Else
        If Lowest > Highest Then
            Dim TempValue = Lowest As Double
            Lowest = Highest
            Highest = TempValue
        End If
        Dim ReturnValue = Lowest + Rnd(0, Highest - Lowest) + Decimal As Double
        If ReturnValue = 0 And PreventZero Then
            Return RandomNumber(Lowest, Highest, DecimalPlaces, PreventZero)
        Else
            Return ReturnValue
        End If
    End If

End Sub

'Prompts the user for a file path and (if applicable) an encryption key, and restores the preferences from the selected file.
'Does nothing if the user interface was not enabled in the Initialize call.
Public Sub Restore

	If prefXCLV.IsInitialized = False Then Return

	Dim fdFile As String
	Dim fdFolder As String = ""

#If B4A
	Dim fd As FileDialog
	fd.FastScroll = True
	fd.FilePath = rp.GetSafeDirDefaultExternal("")
	If fd.Show(prefRestoreText, prefOKtext, prefCancelText, "", Null) <> DialogResponse.POSITIVE Then Return
	fdFile = fd.ChosenName
	fdFolder = fd.FilePath
	If fdFile = "" Then Return
	If File.Exists(fdFolder, fdFile) = False Then Return
#Else If B4J
	Dim fc As FileChooser
	fc.Initialize
	fc.InitialDirectory = dirData
	fc.Title = prefRestoreText
	fdFile = ""
	fdFolder = fc.ShowOpen(callingForm)
	If fdFolder = "" Then Return
	If File.Exists(fdFolder, "") = False Then Return
#End If

	Select Case ReadPrefsFile("", fdFolder, fdFile)
		Case FILE_READ_OK
			WritePrefsFile(prefEncryptionKey, False, "", "")
			jumpToStack.Initialize
			AssembleprefXCLV
		Case FILE_READ_NEED_KEY
			Dim encKey As String = GetEncryptionKey("")
			If encKey = "" Then
				Return
			Else
				Select Case ReadPrefsFile(encKey, fdFolder, fdFile)
					Case FILE_READ_OK
						WritePrefsFile(prefEncryptionKey, False, "", "")
						jumpToStack.Initialize
						AssembleprefXCLV
					Case FILE_READ_WRONG_KEY
						ShowToast(prefWrongEncryptionKey, True)
						Return
					Case Else
						Return
				End Select
			End If
		Case Else
			Return
	End Select

	ShowToast(prefCompletedText, True)

End Sub

'Resets the preferences to the default values that were passed in the AddEntry calls.
'Does nothing if the user interface was not enabled in the Initialize call.
'Does not affect preferences that haven't been setup with 'AddEntry'
Public Sub Reset

	If prefXCLV.IsInitialized = False Then Return

	For Each theEnt As EntryType In allEntries.Values
		prefData.Remove(theEnt.EntryID)
		Try
			If theEnt.defaultValues.IsInitialized Then
				Select Case theEnt.EntryInputType
					Case INPUT_TYPE_INPUTMULTILIST
						Dim currentSelections As List
						currentSelections.Initialize2(theEnt.defaultValues)
						prefData.Put(theEnt.EntryID, currentSelections)
					Case Else
						prefData.Put(theEnt.EntryID, theEnt.defaultValues.Get(0))
				End Select
			End If
		Catch
			Log("") ' No action required
		End Try
	Next

	WritePrefsFile(prefEncryptionKey, False, "", "")
	jumpToStack.Initialize
	AssembleprefXCLV

	ShowToast(prefCompletedText, True)

End Sub

#End Region ' GetSet

#Region FileHandling

Private Sub WritePrefsFile(encKey As String, bkp As Boolean, bkpFolder As String, bkpFile As String)

	Dim theFolder As String
	Dim theFile As String

	If bkp = False Then
		prefData.Put(prefClassID, prefClassVersion)
		prefData.Put(prefClassLastUpdated, DateTime.Now)
		theFolder = prefFolder
		theFile = prefFile
	Else
		If bkpFolder = "" Then
			theFolder = bkpFile
			theFile = ""
		Else
			theFolder = bkpFolder
			theFile = bkpFile
		End If
	End If

	Dim raf As RandomAccessFile
	raf.Initialize(theFolder, theFile, False)

	Select Case encKey
		Case "", Null, "null"
			raf.WriteObject(prefData, False, raf.CurrentPosition) ' Must use RAF and ReadObject/WriteObject because with File.WriteObject, everything would be converted to String, including the List-type objects !
		Case Else
			raf.WriteEncryptedObject(prefData, encKey, raf.CurrentPosition)
	End Select

	raf.Close

End Sub

Private Sub ReadPrefsFile(encryptionKey As String, folderName As String, fileName As String) As Int

	Dim raf As RandomAccessFile
	Dim rafMap As Map
	Dim s As String

	Try
		raf.Initialize(folderName, fileName, True)
	Catch
		Return(FILE_READ_OTHER_ERROR)
	End Try

	If encryptionKey = "" Then
		Try
			rafMap = raf.ReadObject(raf.CurrentPosition) ' Must use RAF and ReadObject/WriteObject because with File.WriteObject, everything would be converted to String, including the List-type objects !
			raf.Close
		Catch
			Dim s As String = LastException.Message.ToLowerCase
			Try
				raf.Close
			Catch
				Log("") ' No action needed
			End Try
			If s.Contains("Data was encrypted. You should provide password.".ToLowerCase) Then
				Return(FILE_READ_NEED_KEY)
			Else
				Return(FILE_READ_OTHER_ERROR)
			End If
		End Try
	Else
		Try
			rafMap = raf.ReadEncryptedObject(encryptionKey, raf.CurrentPosition)
			raf.Close
		Catch
			Try
				raf.Close
			Catch
				Log("") ' No action needed
			End Try
			Return(FILE_READ_WRONG_KEY)
		End Try
	End If

	Try
		s = rafMap.Get(prefClassID)
		Select Case s
			Case "", Null
				Return(FILE_READ_NOT_PREFS_FILE)
		End Select
	Catch
		Return(FILE_READ_NOT_PREFS_FILE)
	End Try

	prefData = rafMap

'	' Cater for files that were created with older versions
'	If prefData.GetDefault(prefClassID, prefClassVersion) = prefClassVersion Then Return(FILE_READ_OK)
'	If prefData.GetDefault(prefClassID, prefClassVersion) >= "00.00.03" Then Return(FILE_READ_OK)
'	' Handle versions < "00.00.03"

	Return(FILE_READ_OK)

End Sub

#End Region ' FileHandling

#Region GeneralPurpose

Private Sub ConvertToBoolean(obj As Object) As Boolean

	Try
		Dim b As Boolean = obj
		Return(b)
	Catch
		Return(False)
	End Try

End Sub

Private Sub ShowToast(theText As String, showLong As Boolean)

#If B4A
	ToastMessageShow(theText, showLong)
#Else If B4J
	cutils.ShowNotification2("", theText, cutils.ICON_INFORMATION, callingForm)
#End If

End Sub

Private Sub ShowMsgBox(theText As String, theTitle As String)

#If B4A
	Msgbox(theText, theTitle)
#Else If B4J
	fx.Msgbox(callingForm, theText, theTitle)
#End If

End Sub

' Only use this Sub to call back in the calling module when an xclv item is clicked
Public Sub xclv_ItemClick(Index As Int, Value As Object)

	Dim ett As EventAndTagType = Value

	Select Case ett.event.ToLowerCase
		Case "category"
			Dim theCat As CategoryType = ett.Tag
			Category_Clicked(theCat, ett.xclvItemIndex)
			Return
		Case "entry"
			Dim theEntry As EntryType = ett.Tag
			Entry_Clicked(theEntry, ett.xclvItemIndex)
	End Select

End Sub

Private Sub PushToJumpToStack(i As Int)

	jumpToStack.InsertAt(0, i)

End Sub

#If B4J
Private Sub IsUnsignedInteger(s As String, allowNullString As Boolean) As Boolean

	' Regex expression based on http://regexlib.com/Search.aspx?k=integer&AspxAutoDetectCookieSupport=1
	If allowNullString And (s = "") Then
		Return True
	Else
		Return IsNumber(s) And Regex.IsMatch("^\d+$", s)
	End If

End Sub

Private Sub IsUnsignedDecimal(s As String, allowNullString As Boolean, decSep As String) As Boolean

	Dim sep As String
	If decSep.Trim = "" Then
		sep = GetDecimalSeparator
	Else
		sep = decSep
	End If

	' Regex expression based on https://community.oracle.com/thread/429716
	If allowNullString And (s = "") Then
		Return True
	Else
		Return IsNumber(s) And Regex.IsMatch("^(\" & sep & "[0-9]+|[0-9]+(\" & sep & "[0-9]*)*)$", s)
	End If

End Sub

Private Sub IsSignedInteger(s As String, allowNullString As Boolean) As Boolean

	' Regex expression based on http://regexlib.com/Search.aspx?k=integer&AspxAutoDetectCookieSupport=1
	If allowNullString And (s = "") Then
		Return True
	Else
		Return IsNumber(s) And Regex.IsMatch("^(\+|-)?\d+$", s)
	End If

End Sub

Private Sub IsSignedDecimal(s As String, allowNullString As Boolean, decSep As String) As Boolean

	Dim sep As String
	If decSep.Trim = "" Then
		sep = GetDecimalSeparator
	Else
		sep = decSep
	End If

	' Regex expression based on https://community.oracle.com/thread/429716
	' combined with http://regexlib.com/Search.aspx?k=integer&AspxAutoDetectCookieSupport=1
	If allowNullString And (s = "") Then
		Return True
	Else
		Return IsNumber(s) And Regex.IsMatch("^(\+|-)?(\" & sep & "[0-9]+|[0-9]+(\" & sep & "[0-9]*)*)$", s)
	End If

End Sub

Private Sub GetDecimalSeparator As String

	Return NumberFormat2(1.1, 1, 1, 1, False).SubString2(1, 2)

End Sub

Private Sub MeasureMultilineTextHeight(Font As Font, Width As Double, Text As String) As Double

	' Code from https://www.b4x.com/android/forum/threads/measure-multiline-text-height.84331

	Dim jo As JavaObject = Me
	Return jo.RunMethod("MeasureMultilineTextHeight", Array(Font, Text, Width))

End Sub

#if Java
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import javafx.scene.text.Font;
import javafx.scene.text.TextBoundsType;
public double MeasureMultilineTextHeight(Font f, String text, double width) throws Exception {
  Method m = Class.forName("com.sun.javafx.scene.control.skin.Utils").getDeclaredMethod("computeTextHeight",
  Font.class, String.class, double.class, TextBoundsType.class);
  m.setAccessible(true);
  return (Double)m.invoke(null, f, text, width, TextBoundsType.LOGICAL);
  }
#End If

Private Sub GetBmpFromFontLabel(txt As String, txtColour As Paint, lbl As Label, bgColour As Paint) As Image

	Dim img As Image

	If lbl.IsInitialized Then
		lbl.Visible = True
		CSSUtils.SetBackgroundColor(lbl, bgColour)
		lbl.Text = txt
		lbl.TextColor = txtColour
		img = lbl.Snapshot
		lbl.Visible = False
		Return img
	Else
		Return Null
	End If

End Sub
#End If ' B4J

#End Region ' GeneralPurpose