﻿B4A=true
Group=Default Group
ModulesStructureVersion=1
Type=StaticCode
Version=7.3
@EndOfDesignText@
'Code module wmUtils

'Subs in this code module will be accessible from all modules.
Sub Process_Globals

	'These global variables will be declared once when the application starts.
	'These variables can be accessed from all modules.

End Sub

Sub wmHasFeature(feature As String) As Boolean

	' Code from https://www.b4x.com/android/forum/threads/phoney-questions.23158/#post-134429

	Dim r As Reflector
	r.Target = r.GetContext
	r.Target = r.RunMethod("getPackageManager")
	Return r.RunMethod2("hasSystemFeature", feature, "java.lang.String")

End Sub

Sub wmListToOneString(arr As List, sep As String) As String

	Dim s As String = ""
	Dim i As Int

	If arr = Null Then Return ""

	For i = 0 To (arr.Size - 1)
		If s <> "" Then s = s & sep
		s = s & arr.get(i)
	Next

	Return s

End Sub

Sub wmLogWithTimestamp(txt As String)

	Log(DateTime.Date(DateTime.Now) & " " & DateTime.Time(DateTime.Now) & " : " & txt)

End Sub

Public Sub wmRandomNumber(Lowest As Double, Highest As Double, DecimalPlaces As Int, PreventZero As Boolean) As Double

	' Code from https://www.b4x.com/android/forum/threads/randomnumber-the-most-powerful-number-generator-b4a-b4j-b4i.58456/
	' Returns a random double between Lowest (inclusive) and Highest (exclusive), with n decimal places.

	Lowest = Round(Lowest)
	Highest = Round(Highest)
	Dim Decimal As Double
	If DecimalPlaces > 0 Then Decimal = (Rnd(0, Power(10, DecimalPlaces))) / Power(10, DecimalPlaces)
	If Lowest = Highest Then
		Return Lowest
	Else
		If Lowest > Highest Then
			Dim TempValue = Lowest As Double
			Lowest = Highest
			Highest = TempValue
		End If
		Dim ReturnValue = Lowest + Rnd(0, Highest - Lowest) + Decimal As Double
		If ReturnValue = 0 And PreventZero Then
			Return wmRandomNumber(Lowest, Highest, DecimalPlaces, PreventZero)
		Else
			Return ReturnValue
		End If
	End If

End Sub

Public Sub wmIsUnsignedInteger(s As String, allowNullString As Boolean) As Boolean

	' Regex expression based on http://regexlib.com/Search.aspx?k=integer&AspxAutoDetectCookieSupport=1
	If allowNullString And (s = "") Then
		Return True
	Else
		Return IsNumber(s) And Regex.IsMatch("^\d+$", s)
	End If

End Sub

Public Sub wmIsUnsignedDecimal(s As String, allowNullString As Boolean, decSep As String) As Boolean

	Dim sep As String
	If decSep.Trim = "" Then
		sep = wmGetDecimalSeparator
	Else
		sep = decSep
	End If

	' Regex expression based on https://community.oracle.com/thread/429716
	If allowNullString And (s = "") Then
		Return True
	Else
		Return IsNumber(s) And Regex.IsMatch("^(\" & sep & "[0-9]+|[0-9]+(\" & sep & "[0-9]*)*)$", s)
	End If

End Sub

Public Sub wmIsSignedInteger(s As String, allowNullString As Boolean) As Boolean

	' Regex expression based on http://regexlib.com/Search.aspx?k=integer&AspxAutoDetectCookieSupport=1
	If allowNullString And (s = "") Then
		Return True
	Else
		Return IsNumber(s) And Regex.IsMatch("^(\+|-)?\d+$", s)
	End If

End Sub

Public Sub wmIsSignedDecimal(s As String, allowNullString As Boolean, decSep As String) As Boolean

	Dim sep As String
	If decSep.Trim = "" Then
		sep = wmGetDecimalSeparator
	Else
		sep = decSep
	End If

	' Regex expression based on https://community.oracle.com/thread/429716
	' combined with http://regexlib.com/Search.aspx?k=integer&AspxAutoDetectCookieSupport=1
	If allowNullString And (s = "") Then
		Return True
	Else
		Return IsNumber(s) And Regex.IsMatch("^(\+|-)?(\" & sep & "[0-9]+|[0-9]+(\" & sep & "[0-9]*)*)$", s)
	End If

End Sub

Public Sub wmGetDecimalSeparator As String

	Return NumberFormat2(1.1, 1, 1, 1, False).SubString2(1, 2)

End Sub

Public Sub wmZeroIfEmptyTrimmedString(s As String) As String

	If s.Trim = "" Then
		Return "0"
	Else
		Return s.Trim
	End If

End Sub

Public Sub wmUnescapeUnicode(s As String) As String

	' Code from https://www.b4x.com/android/forum/threads/when-getting-query-results-i-cant-read-hebrew.27461/#post-159533

	Dim i As Int
	Dim sb As StringBuilder

	sb.Initialize
	Do While i < s.Length
		Dim c As Char = s.CharAt(i)
		If c = "\" And i < s.Length - 1 And s.CharAt(i + 1) = "u" Then
			Dim unicode As StringBuilder
			unicode.Initialize
			i = i + 2
			Do While i < s.Length
				Dim cc As String = s.CharAt(i)
				Dim n As Int = Asc(cc.ToLowerCase)
				If (n >= Asc("0") And n <= Asc("9")) Or (n >= Asc("a") And n <= Asc("z")) Then
					unicode.Append(s.CharAt(i))
				Else
					i = i - 1
					Exit
				End If
				i = i + 1
			Loop
			sb.Append(Chr(Bit.ParseInt(unicode.ToString, 16)))
		Else
			sb.Append(c)
		End If
		i = i + 1
	Loop

	Return sb.ToString

End Sub

' This sub comes from http://www.basic4ppc.com/android/forum/threads/how-to-strip-invalid-xml-characters.18219/#post105025
' We are removing utf-8 characters that may not appear in well-formed XML documents.
' see W3C - Extensible Markup Language (XML) 1.1 (Second Edition) for valid character ranges
' http://www.w3.org/TR/2006/REC-xml11-20060816/#charsets
Sub wmXmlCharacterWhitelist(in_string As String) As String
	If( in_string == Null Or in_string == "") Then
		Return Null
	End If

	Dim sbOutput As StringBuilder : sbOutput.Initialize ' stirng builder to keep our valid characters
	Dim ch As Char ' A char is 16 bits in Java (and is also the only unsigned type!!)
	
	For i  = 0 To in_string.Length - 1
		ch = in_string.CharAt(i)
		Dim ch_number As Int : ch_number = Asc(ch) ' chr asc
		
		' any Unicode character, excluding the surrogate blocks, FFFE, and FFFF.
		If ((ch_number >= 0x0001 And ch_number <= 0xD7FF ) Or _
                (ch_number >= 0xE000 And ch_number <= 0xFFFD ) Or _
            (ch_number >= 0x10000 And ch_number <= 0x10FFFF) Or _
                ch_number == 0x0009 Or _
                ch_number == 0x000A Or  _
                ch_number == 0x000D ) Then
            
			' Restricted Chars
			If ((ch_number < 0x1  Or ch_number > 0x8) And _
               (ch_number < 0xB  Or ch_number > 0xC)      And _
               (ch_number < 0xE  Or ch_number > 0x1F)   And _
               (ch_number < 0x7F Or ch_number > 0x84)   And _
               (ch_number < 0x86 Or ch_number > 0x9F)) Then
				sbOutput.Append(ch)
			End If
		End If
	Next

	Return sbOutput.ToString

End Sub

Sub wmGetFromMapAsInt(m As Map, key As String, deflt As Object) As Int

	Try
		Return m.GetDefault(key, deflt)
	Catch
		Return deflt
	End Try

End Sub

Sub wmGetFromMapAsString(m As Map, key As String, deflt As Object) As String

	Try
		Return m.GetDefault(key, deflt)
	Catch
		Return deflt
	End Try

End Sub

Sub wmGetFromMapAsBoolean(m As Map, key As String, deflt As Object) As Boolean

	Try
		Dim s As String = m.GetDefault(key, deflt)
		Select Case s.Trim.ToLowerCase
			Case "true"
				Return True
			Case "false"
				Return False
			Case Else
				Return deflt
		End Select
	Catch
		Return deflt
	End Try

End Sub

Sub wmParseDelimitedString(theString As String, preamble As String, postamble As String) As String

	Dim reslt As String = theString
	Dim pre As String = wmEscapeForRegex(preamble)
	Dim post As String = wmEscapeForRegex(postamble)
	Dim splt1() As String
	Dim splt2() As String

	Try
		If pre = "" Then
			If post = "" Then
				reslt = theString
			Else
				splt2 = Regex.Split2(post, Regex.CASE_INSENSITIVE, theString)
				reslt = splt2(0)
			End If
		Else
			splt1 = Regex.Split2(pre, Regex.CASE_INSENSITIVE, theString)
			If post = "" Then
				reslt = splt1(1)
			Else
				splt2 = Regex.Split2(post, Regex.CASE_INSENSITIVE, splt1(1))
				reslt = splt2(0)
			End If
		End If
	Catch
		reslt = "" ' Just getting rid of a compiler warning here
#If DEBUG
		Log("wmParseDelimitedString: " & LastException)
#End If
	End Try

	Return reslt

End Sub

Sub wmEscapeForRegex(theString As String) As String

	Dim lst As List = Array As String("\", "^", "$", ".", "|", "?", "*", "+", "(", ")", "[", "{") ' As per https://www.regular-expressions.info/characters.html
	Dim reslt As String = theString

	For Each s As String In lst
		reslt = reslt.Replace(s, "\" & s)
	Next

	Return reslt

End Sub

Sub wmParseDelimitedString2(theString As String, preamble As String, postamble As String, ignoreCase As Boolean) As String

	If (preamble = "") And (postamble = "") Then Return theString

	Dim reslt As String = ""
	Dim theStringLow As String
	Dim preLow As String
	Dim postLow As String
	Dim pos As Int

	If ignoreCase Then
		theStringLow = theString.ToLowerCase
		preLow = preamble.ToLowerCase
		postLow = postamble.ToLowerCase
	Else
		theStringLow = theString
		preLow = preamble
		postLow = postamble
	End If

	Try
		If preamble = "" Then
			pos = theStringLow.IndexOf(postLow)
			If pos >= 0 Then reslt = theString.SubString2(0, pos)
		Else
			pos = theStringLow.IndexOf(preLow)
			If pos >= 0 Then
				Dim theSubString As String = theString.SubString(pos + preLow.Length)
				Dim theSubStringLow As String
				If ignoreCase Then
					theSubStringLow = theSubString.ToLowerCase
				Else
					theSubStringLow = theSubString
				End If
				If postLow = "" Then
					reslt = theSubString
				Else
					pos = theSubStringLow.IndexOf(postLow)
					If pos >= 0 Then reslt = theSubString.SubString2(0, pos)
				End If
			End If
		End If
	Catch
		reslt = "" ' Just getting rid of a compiler warning here
#If DEBUG
		Log("wmParseDelimitedString2: " & LastException)
#End If
	End Try

	Return reslt

End Sub

#If Not(NON_UI)

Public Sub wmGetBottom(Vw As B4XView) As Int

	' Code from https://www.b4x.com/android/forum/threads/xui-source-code-viewpos.92820/#post-587225
	Return Vw.Top + Vw.Height - 1

End Sub

Public Sub wmGetRight(Vw As B4XView) As Int

	' Code from https://www.b4x.com/android/forum/threads/xui-source-code-viewpos.92820/#post-587225
	Return Vw.Left + Vw.Width - 1

End Sub

Sub wmSetRight(Vw As B4XView, Right As Int)

	' Code from https://www.b4x.com/android/forum/threads/xui-source-code-viewpos.92820/#post-587225
	Vw.Left = Right - Vw.Width

End Sub

Public Sub wmSetBottom(Vw As B4XView, Bottom As Int)

	' Code from https://www.b4x.com/android/forum/threads/xui-source-code-viewpos.92820/#post-587225
	Vw.Top = Bottom - Vw.Height

End Sub

#End If ' Not(NON_UI)

#If B4A
Sub wmAddBorderToPanel(thePanel As Panel, thePanelColor As Int, theRadius As Int, theBorderWidth As Int, theBorderColor As Int)

	' Based on http://www.b4x.com/android/forum/threads/panel-borders-get-lost-when-changing-color.49078/
	If theBorderWidth > 0 Then
		Dim cdw As ColorDrawable
		cdw.Initialize2(thePanelColor, theRadius, theBorderWidth, theBorderColor)
		thePanel.Background = cdw
	End If

End Sub

Sub wmAddBorderToView(theView As View, theViewColor As Int, theRadius As Int, theBorderWidth As Int, theBorderColor As Int)

	' Based on http://www.b4x.com/android/forum/threads/panel-borders-get-lost-when-changing-color.49078/
	If theBorderWidth > 0 Then
		Dim cdw As ColorDrawable
		cdw.Initialize2(theViewColor, theRadius, theBorderWidth, theBorderColor)
		theView.Background = cdw
		If theRadius <> 0 Then wmSetCornerRadii(theView, theRadius, theRadius, theRadius, theRadius, theRadius, theRadius, theRadius, theRadius)
	End If

End Sub

Sub wmSetCornerRadii(v As View, Rx_TopLeft As Float, Ry_TopLeft As Float, Rx_TopRight As Float, Ry_TopRight As Float, Rx_BottomRight As Float, Ry_BottomRight As Float, Rx_BottomLeft As Float, Ry_BottomLeft As Float)

	' Based on https://www.b4x.com/android/forum/threads/gradientdrawable-with-different-corner-radius.51475/#post-322392

	Dim jo As JavaObject = v.Background
	If v.Background Is ColorDrawable Or v.Background Is GradientDrawable Then
		jo.RunMethod("setCornerRadii", Array As Object(Array As Float(Rx_TopLeft, Ry_TopLeft, Rx_TopRight, Ry_TopRight, Rx_BottomRight, Ry_BottomRight, Rx_BottomLeft, Ry_BottomLeft)))
	End If

End Sub

Sub wmGetAppString(theTxt As String) As String

	Try
#If nolocalizator
		Return (theTxt)
#Else
		Return Starter.loc.Localize(theTxt)
#End If
	Catch
		Return "Missing resource string: " & theTxt
	End Try

End Sub

Sub wmScrollViewSetup(theSv As View, panelColor As Int, svLeft As Int, svWidth As Int, gradientOrientation As String, gradientColour1 As Int, gradientColour2 As Int)

	If theSv Is ScrollView Then
		Dim sv As ScrollView = theSv
		sv.Panel.RemoveAllViews
		sv.Panel.Height = 0
		sv.Panel.Color = panelColor
		sv.Left = svLeft
		sv.Width = svWidth
		sv.Panel.Width = svWidth
		If gradientOrientation <> "" Then
			Dim gd1 As GradientDrawable
			Dim cols1(2) As Int
			cols1(0) = gradientColour1
			cols1(1) = gradientColour2
			gd1.Initialize(gradientOrientation, cols1)
			gd1.CornerRadius=0
			sv.Panel.Background = gd1
			'sv.Background = gd1
		End If
		If sv.IsInitialized = False Then sv.Initialize(10dip)
	Else
		Dim hsv As HorizontalScrollView = theSv
		hsv.Panel.RemoveAllViews
		hsv.Panel.Height = 0
		hsv.Panel.Color = panelColor
		hsv.Left = svLeft
		hsv.Width = svWidth
		hsv.Panel.Width = svWidth
		If gradientOrientation <> "" Then
			Dim gd1 As GradientDrawable
			Dim cols1(2) As Int
			cols1(0) = gradientColour1
			cols1(1) = gradientColour2
			gd1.Initialize(gradientOrientation, cols1)
			gd1.CornerRadius=0
			hsv.Panel.Background = gd1
			'hsv.Background = gd1
		End If
		If hsv.IsInitialized = False Then hsv.Initialize(10dip, "")
	End If

End Sub

Sub wmTextToBitmapMaterialFont28(s As String, txtColor As Int, lblMaterial As Label) As Bitmap

	Return wmTextToBitmap(s, 28, True, txtColor, lblMaterial, Null)

End Sub

Sub wmTextToBitmapMaterial(s As String, FontSize As Float, txtColor As Int, lblMaterial As Label) As Bitmap

	Return wmTextToBitmap(s, FontSize, True, txtColor, lblMaterial, Null)

End Sub

Sub wmTextToBitmapFontAwesomeFont28(s As String, txtColor As Int, lblFontAwesome As Label) As Bitmap

	Return wmTextToBitmap(s, 28, False, txtColor, Null, lblFontAwesome)

End Sub

Sub wmTextToBitmapFontAwesome(s As String, FontSize As Float, txtColor As Int, lblFontAwesome As Label) As Bitmap

	Return wmTextToBitmap(s, FontSize, False, txtColor, Null, lblFontAwesome)

End Sub

Sub wmTextToBitmap(s As String, FontSize As Float, useFontMaterial As Boolean, txtColor As Int, lblMaterial As Label, lblFontAwesome As Label) As Bitmap

	Return wmTextToBitmapDo(s, FontSize, useFontMaterial, txtColor, lblMaterial, lblFontAwesome, 32dip, 32dip)

End Sub

Sub wmTextToBitmap2(s As String, FontSize As Float, useFontMaterial As Boolean, txtColor As Int, lblMaterial As Label, lblFontAwesome As Label, bmpWidth As Int, bmpHeight As Int) As Bitmap

	Return wmTextToBitmapDo(s, FontSize, useFontMaterial, txtColor, lblMaterial, lblFontAwesome, bmpWidth, bmpHeight)

End Sub

Sub wmTextToBitmapDo(s As String, FontSize As Float, useFontMaterial As Boolean, txtColor As Int, lblMaterial As Label, lblFontAwesome As Label, bmpWidth As Int, bmpHeight As Int) As Bitmap

	' Useful to get the drawable from FontAwesome or Material Icons
	' Based on https://www.b4x.com/android/forum/threads/new-feature-fontawesome-and-material-icons.73646
	' To search in these icon sets:
	' - https://fontawesome.io/icons
	' - https://material.io/icons
	' Alternatively, drawable IDs and images can be found at:
	' http://developer.android.com/reference/android/R.drawable.html
	' App AndroidDrawablesViewerWM

	Dim bmp As Bitmap
	bmp.InitializeMutable(bmpWidth, bmpHeight)
	Dim cvs As Canvas
	cvs.Initialize2(bmp)

	If useFontMaterial Then
		Dim h As Double = cvs.MeasureStringHeight(s, lblMaterial.Typeface, FontSize)
		cvs.DrawText(s, bmp.Width / 2, bmp.Height / 2 + h / 2, lblMaterial.Typeface, FontSize, txtColor, "CENTER")
	Else
		Dim h As Double = cvs.MeasureStringHeight(s, lblFontAwesome.Typeface, FontSize)
		cvs.DrawText(s, bmp.Width / 2, bmp.Height / 2 + h / 2, lblFontAwesome.Typeface, FontSize, txtColor, "CENTER")
	End If

	Return bmp

End Sub

Sub wmTextToBitmapCenterMaterialFont28(s As String, txtXuiColor As Int) As B4XBitmap

	Return wmTextToBitmapCenter(s, 28, True, txtXuiColor)

End Sub

Sub wmTextToBitmapCenterMaterial(s As String, FontSize As Float, txtXuiColor As Int) As Bitmap

	Return wmTextToBitmapCenter(s, FontSize, True, txtXuiColor)

End Sub

Sub wmTextToBitmapCenterFontAwesomeFont28(s As String, txtXuiColor As Int) As Bitmap

	Return wmTextToBitmapCenter(s, 28, False, txtXuiColor)

End Sub

Sub wmTextToBitmapCenterFontAwesome(s As String, FontSize As Float, txtXuiColor As Int) As Bitmap

	Return wmTextToBitmapCenter(s, FontSize, False, txtXuiColor)

End Sub

Sub wmTextToBitmapCenter(s As String, FontSize As Float, useFontMaterial As Boolean, txtXuiColor As Int) As Bitmap

	Return wmTextToBitmapCenterDo(s, FontSize, useFontMaterial, txtXuiColor, 32dip, 32dip)

End Sub

Sub wmTextToBitmapCenter2(s As String, FontSize As Float, useFontMaterial As Boolean, txtXuiColor As Int, bmpWidth As Int, bmpHeight As Int) As Bitmap

	Return wmTextToBitmapCenterDo(s, FontSize, useFontMaterial, txtXuiColor, bmpWidth, bmpHeight)

End Sub

Sub wmTextToBitmapCenterDo(s As String, fontSize As Float, useFontMaterial As Boolean, txtXuiColor As Int, bmpWidth As Int, bmpHeight As Int) As B4XBitmap

	' Based on https://www.b4x.com/android/forum/threads/fontawesome-to-bitmap.95155/
	' Requires XUI library
	' This one also centers the icon vertically

	Dim xui As XUI
	Dim p As Panel = xui.CreatePanel("")
	p.SetLayoutAnimated(0, 0, 0, bmpWidth, bmpHeight)
	Dim cvs1 As B4XCanvas
	cvs1.Initialize(p)
	Dim fnt As B4XFont
	If useFontMaterial Then
		fnt = xui.CreateFont(Typeface.MATERIALICONS, fontSize)
	Else
		fnt = xui.CreateFont(Typeface.FONTAWESOME, fontSize)
	End If
	Dim r As B4XRect = cvs1.MeasureText(s, fnt)
	Dim BaseLine As Int = cvs1.TargetRect.CenterY - r.Height / 2 - r.Top
	cvs1.DrawText(s, cvs1.TargetRect.CenterX, BaseLine, fnt, txtXuiColor, "CENTER")
	Dim b As B4XBitmap = cvs1.CreateBitmap
	cvs1.Release
	Return b

End Sub

Public Sub wmAddCSmenuItem(act As Activity, title As String, eventName As String, textColour As Int, fontAwesomeChr As String, materialIconsChr As String)

	Dim cs As CSBuilder

	If fontAwesomeChr<>"" Then
		cs.Initialize.Typeface(Typeface.FONTAWESOME).Append(fontAwesomeChr).Pop.Color(textColour).Append(" " & title).PopAll
	Else
		cs.Initialize.Typeface(Typeface.MATERIALICONS).Append(materialIconsChr).Pop.Color(textColour).Append(" " & title).PopAll
	End If

	act.AddMenuItem(cs, eventName)

End Sub

Sub wmStringToCSBuilder(s As String) As CSBuilder

	Dim cs As CSBuilder
	Return cs.Initialize.Append(s)

End Sub

Sub wmCloseAllActivities

	' Code from https://www.b4x.com/android/forum/threads/exitapplication-vs-activity-finish.79548/#post-503759
	' If you are targeting Android 4.1+ (API 16+) then you can use this code to close all activities

	Dim jo As JavaObject
	jo.InitializeContext
	jo.RunMethod("finishAffinity", Null)

End Sub

Sub wmDebugLogWrite(txt As String)

	Log("wmDebugLog: " & txt)
	wmDebugLogWrite2(txt, True)

End Sub

Sub wmDebugLogReset

	wmDebugLogWrite2("", False)

End Sub

Private Sub wmDebugLogWrite2(txt As String, mustappend As Boolean)

	Dim logfile As TextWriter
	Dim logfileName As String = File.Combine(File.DirInternal, Application.LabelName & ".debuglog.txt")

	Try
		logfile.Initialize(File.OpenOutput(logfileName, "", mustappend))
		logfile.WriteLine("==================================")
		logfile.WriteLine(DateTime.Date(DateTime.Now) & " " & DateTime.Time(DateTime.Now) & " : " & txt)
		logfile.Close
	Catch
		Log("===========================")
		Log("Failed to write to logfile: " & LastException)
		Log("Logfile is: " & logfileName)
		Log("Was attempting to write: " & txt)
		Log("===========================")
	End Try

End Sub

Sub wmDebugLogShow(title As String)

	Dim content As String = ""
	Dim sz As Int = -1
	Dim strSz As String

	If File.Exists(File.DirInternal, Application.LabelName & ".debuglog.txt") = False Then
		Msgbox("", title)
		Return
	End If

	If title = "" Then title = "wmDebugLog"

	Try
		sz = File.Size(File.DirInternal, Application.LabelName & ".debuglog.txt")
		content = File.ReadString(File.DirInternal, Application.LabelName & ".debuglog.txt")
	Catch
		content = "wmDebugLogShow: " & LastException
	End Try

	If sz < 1000 Then
		strSz = sz
	Else If sz < 1000000 Then
		strSz = (sz / 1000) & "kb"
	Else If sz < 1000000000 Then
		strSz = (sz / 1000000) & "mb"
	Else
		strSz = (sz / 1000000000) & "gb"
	End If

	content = "Debug log file size: " & strSz & CRLF & content & "==================================" & CRLF & "Debug log file size: " & strSz

	Try
		Dim clip As BClipboard
		clip.clrText
		clip.setText(content)
		title = title & " - copied to Clipboard"
	Catch
		Log("wmDebugLogShow: " & LastException)
	End Try

	Try
		Msgbox(content, title)
	Catch
		Log("wmDebugLogShow: " & LastException)
	End Try

End Sub

'Sub ShowWmDebugLogFile_Click
'
'	' Based on https://www.b4x.com/android/forum/threads/share-txt-files.73166/#content
'
'	Dim logfileName As String = File.Combine(File.DirRootExternal, Application.LabelName & ".debuglog.txt")
'	Dim rp As RuntimePermissions
'	Dim shared As String = rp.GetSafeDirDefaultExternal("shared")
'
'	If File.Exists(logfileName, "") = False Then
'		Msgbox("Logfile does not exist:" & CRLF & logfileName, Application.LabelName)
'		Return
'	End If
'
'	'copy the file to the shared folder
'	File.Copy(logfileName, "", shared, Application.LabelName & ".debuglog.txt")
'	Dim in As Intent
'	in.Initialize(in.ACTION_SEND, "")
'	in.SetType("text/plain") 'it is not related to the file itself.
'	in.PutExtra("android.intent.extra.STREAM",  CreateFileProviderUri(shared, "1.bal"))
'	in.Flags = 1
'	StartActivity(in)
'
'End Sub
'
'Sub CreateFileProviderUri (Dir As String, FileName As String) As Object
'
'	' Code from https://www.b4x.com/android/forum/threads/share-txt-files.73166/#content
'	Dim FileProvider As JavaObject
'	Dim context As JavaObject
'	context.InitializeContext
'	FileProvider.InitializeStatic("android.support.v4.content.FileProvider")
'	Dim f As JavaObject
'	f.InitializeNewInstance("java.io.File", Array(Dir, FileName))
'	Return FileProvider.RunMethod("getUriForFile", Array(context, Application.PackageName & ".provider", f))
'End Sub

#End If 'B4A

#If B4J And (Not (NON_UI))

Sub Beep

	' Code from https://www.b4x.com/android/forum/threads/beep.35661/#post-209569
	Dim jo As JavaObject
	jo.InitializeStatic("java.awt.Toolkit")
	Dim toolkit As JavaObject = jo.RunMethod("getDefaultToolkit", Null)
	toolkit.RunMethod("beep", Null)

End Sub

#End If ' B4J And (Not (NON_UI))